from fiscalsim_us.model_api import *


class ar_income_tax_before_non_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arkansas income tax before non refundable credits"
    unit = USD
    definition_period = YEAR
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2023_Final_AR1000ES.pdf"
    defined_for = StateCode.AR

    def formula(tax_unit, period, parameters):
        rate = parameters(period).gov.states.ar.tax.income.rates.main
        taxable_income = tax_unit("ar_taxable_income", period)

        return rate.calc(taxable_income)



import pandas as pd

# Per-ZCTA population dataset
# ACS 5-year estimates, download URL: https://data.census.gov/cedsci/table?q=DP05%3A%20ACS%20DEMOGRAPHIC%20AND%20HOUSING%20ESTIMATES&g=0100000US%248600000&tid=ACSDP5Y2020.DP05

zcta = pd.read_csv(
    "zcta_2020_population.csv",
    low_memory=False,
    usecols=["DP05_0001E", "NAME"],
    skiprows=1,
    header=0,
    names=["population", "zcta"],
)
zcta.zcta = zcta.zcta.apply(lambda x: x.split(" ")[1])
zcta.zcta = zcta.zcta.astype(int)
zcta.population = zcta.population.astype(int)

# ZCTA-county dataset
# 2020 ZCTA to County Relationship File, download URL: https://www.census.gov/geographies/reference-files/time-series/geo/relationship-files.html#zcta
zcta_to_county = pd.read_csv(
    "zcta_2020_to_county_2020.csv",
    delimiter="|",
    usecols=["GEOID_ZCTA5_20", "NAMELSAD_COUNTY_20"],
    names=["zcta", "county"],
)
zcta_to_county = zcta_to_county.dropna()
zcta_to_county.zcta = zcta_to_county.zcta.astype(int)
# Some ZCTAs have more than one county - select a random one
zcta_to_county = zcta_to_county.groupby("zcta").apply(lambda x: x.sample(1))

# ZIP code-ZCTA dataset
# Download URL: https://udsmapper.org/zip-code-to-zcta-crosswalk/
zip_code = pd.read_csv(
    "zip_code_to_zcta.csv",
    usecols=["ZIP_CODE", "ZCTA", "STATE"],
    names=["zip_code", "zcta", "state"],
)
zip_code.zip_code = zip_code.zip_code.astype(int)
zip_code = zip_code[zip_code.zcta != "No ZCTA"]
zip_code.zcta = zip_code.zcta.astype(int)
zip_code = zip_code[zip_code.zcta.isin(zcta.zcta)]
zip_code = zip_code[zip_code.zcta.isin(zcta_to_county.zcta)]

# ZCTAs have multiple ZIP codes - split each ZCTA population equally into its component ZIP codes
zip_code["population"] = (
    zcta.set_index("zcta").population[zip_code.zcta].values
    / zip_code.groupby("zcta").zip_code.count()[zip_code.zcta].values
)
zip_code["county"] = (
    zcta_to_county.set_index("zcta").county[zip_code.zcta].values
)
zip_code.to_csv("zip_codes.csv", compression="gzip")



from fiscalsim_us.model_api import *


class social_security_taxable_self_employment_income(Variable):
    value_type = float
    entity = Person
    label = "Taxable self-employment income for computing Social Security tax"
    definition_period = YEAR
    unit = USD

    def formula(person, period, parameters):
        ss_cap = parameters(period).gov.irs.payroll.social_security.cap
        # Deduct SS payroll taxable wages and salaries.
        return min_(
            person("taxable_self_employment_income", period),
            ss_cap - person("taxable_earnings_for_social_security", period),
        )



from fiscalsim_us.model_api import *


class va_fixed_date_conformity_subtractions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Va fixed date conformity subtractions https://www.tax.virginia.gov/sites/default/files/taxforms/individual-income-tax/2021/schedule-adj-2021.pdf"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.VA



from fiscalsim_us.model_api import *


class mt_charitable_cash_deduction(Variable):
    """
    Line 11 on Montana itemized deductions schedule
    """

    value_type = float
    entity = TaxUnit
    label = "Montana charitable cash donation deduction"
    unit = USD
    documentation = (
        "Montana deduction from taxable income for charitable cash donations."
    )
    definition_period = YEAR
    defined_for = StateCode.MT

    def formula(tax_unit, period, parameters):
        cash_donations = add(tax_unit, period, ["charitable_cash_donations"])

        return cash_donations



from fiscalsim_us.model_api import *


class va_witholding(Variable):
    value_type = float
    entity = TaxUnit
    label = "va tax witheld from the tax year line 19a"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.VA



from fiscalsim_us.model_api import *


class flat_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "Flat tax"
    unit = USD
    documentation = "Flat income tax on federal AGI."
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        agi = tax_unit("positive_agi", period)
        p = parameters(period).gov.contrib.ubi_center.flat_tax
        return p.rate * agi



from fiscalsim_us.model_api import *


class is_child_of_tax_head(Variable):
    value_type = bool
    entity = Person
    label = "Is a child"
    definition_period = YEAR

    def formula(person, period, parameters):
        is_child_of_tax_head = person("is_child", period)
        return is_child_of_tax_head



from fiscalsim_us.model_api import *


class ctc_qualifying_child(Variable):
    value_type = bool
    entity = Person
    label = "CTC-qualifying child"
    documentation = "Child qualifies for the Child Tax Credit"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/24#c"

    def formula(person, period, parameters):
        return person("ctc_child_individual_maximum", period) > 0



from fiscalsim_us.model_api import *


class dwks6(Variable):
    value_type = float
    entity = TaxUnit
    label = "DWKS6"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        dwks2 = add(tax_unit, period, ["qualified_dividend_income"])
        dwks3 = tax_unit("investment_income_form_4952", period)
        # dwks4 always assumed to be zero
        dwks5 = max_(0, dwks3)
        return max_(0, dwks2 - dwks5)



from fiscalsim_us.model_api import *


class va_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "line 26 on the va tax return"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.VA

    def formula(tax_unit, period, parameters):
        line_19 = tax_unit("va_witholding", period)

        line_19b = tax_unit("va_spouse_witholding", period)

        line_20 = tax_unit("va_estimated_tax_payments_for_tax_year", period)

        line_21 = tax_unit(
            "va_amount_of_prior_year_overpayment_applied_to_current_year",
            period,
        )

        line_22 = tax_unit("va_extension_payments", period)

        line_23 = tax_unit("va_tax_credit_for_low_income_individuals", period)

        line_24 = tax_unit("va_credits_from_enclosed_schedule_cr", period)

        return (
            line_19
            + line_19b
            + line_20
            + line_21
            + line_22
            + line_24
            + line_23
        )



from fiscalsim_us.model_api import *


class prescription_expense(Variable):
    value_type = float
    entity = Person
    label = "Prescription expenses"
    unit = USD
    definition_period = YEAR



from fiscalsim_us.model_api import *


class ut_income_tax(Variable):
    """
    Line 40 of Utah 2022 Individual Income Tax return form TC-40.
    """

    value_type = float
    entity = TaxUnit
    label = "UT income tax after refundable credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.UT
    adds = ["ut_income_tax_before_refundable_credits"]
    subtracts = ["ut_refundable_credits"]



from fiscalsim_us.model_api import *


class me_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maine EITC"
    unit = USD
    definition_period = YEAR
    reference = "https://www.mainelegislature.org/legis/statutes/36/title36sec5219-S.html"
    defined_for = StateCode.ME

    def formula(tax_unit, period, parameters):
        # First get their federal EITC.
        eitc = tax_unit("earned_income_tax_credit", period)

        # Then get the ME EITC part of the parameter tree.
        p = parameters(period).gov.states.me.tax.income.credits.eitc

        # Determine applicable percentage of federal EITC.
        # Depends on whether or not they have at least one qualifying child.
        percentage = where(
            tax_unit("eitc_child_count", period) > 0,
            p.rate.with_qualifying_child,
            p.rate.no_qualifying_child,
        )

        # Return the net Maine EITC.
        return eitc * percentage



from fiscalsim_us.model_api import *


class mt_exemptions_count(Variable):
    value_type = int
    entity = TaxUnit
    label = "Number of Montana exemptions"
    definition_period = YEAR
    defined_for = StateCode.MT

    adds = [
        "head_spouse_count",
        "blind_head",
        "blind_spouse",
        "mt_dependent_exemption_count",
        "mt_aged_exemption_count",
    ]



from fiscalsim_us.model_api import *


class charitable_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Charitable deduction"
    unit = USD
    documentation = "Deduction from taxable income for charitable donations."
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/170"

    def formula(tax_unit, period, parameters):
        cash_donations = add(tax_unit, period, ["charitable_cash_donations"])
        non_cash_donations = add(
            tax_unit, period, ["charitable_non_cash_donations"]
        )
        positive_agi = tax_unit("positive_agi", period)
        ceiling = parameters(
            period
        ).gov.irs.deductions.itemized.charity.ceiling
        capped_non_cash_donations = min_(
            non_cash_donations, ceiling.non_cash * positive_agi
        )
        return min_(
            capped_non_cash_donations + cash_donations,
            ceiling.all * positive_agi,
        )



from fiscalsim_us.model_api import *


class ny_itemized_deductions_max(Variable):
    value_type = float
    entity = TaxUnit
    label = "NY uncapped itemized deductions"
    unit = USD
    definition_period = YEAR
    reference = "https://www.nysenate.gov/legislation/laws/TAX/615"
    defined_for = StateCode.NY

    def formula(tax_unit, period, parameters):
        itm_deds_less_salt = tax_unit("itemized_deductions_less_salt", period)
        capped_property_taxes = tax_unit("capped_property_taxes", period)
        p = parameters(period).gov.states.ny.tax.income
        capped_tuition = min_(
            p.deductions.itemized.college_tuition_max,
            add(tax_unit, period, ["qualified_tuition_expenses"]),
        )
        return itm_deds_less_salt + capped_property_taxes + capped_tuition



from fiscalsim_us.model_api import *


class tanf_initial_employment_deduction(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = YEAR
    label = "TANF IED (Initial Employment Deduction)"
    documentation = "The amount deducted from the countable earnings of a TANF application when calculating initial eligibility."
    unit = USD

    def formula_2022(spm_unit, period, parameters):
        family_size = spm_unit("spm_unit_size", period)
        state = spm_unit.household("state_code_str", period)
        ied = parameters(
            period
        ).gov.hhs.tanf.cash.eligibility.initial.deductions.employment
        per_earner = ied.earner[state]
        per_household = ied.household[state][family_size]
        earners = spm_unit.members("market_income", period) > 0
        earner_count = spm_unit.sum(earners)
        return (per_household + (per_earner * earner_count)) * MONTHS_IN_YEAR



from fiscalsim_us.model_api import *


class mt_long_term_care_premiums_deduction(Variable):
    """
    Line 3 on itemized deductions schedule
    Long-term core insurance premiums not deducted elsewhere on your return
    """

    value_type = float
    entity = TaxUnit
    label = "Montana long-term care insurance premiums deduction"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT



from fiscalsim_us.model_api import *


class mt_charitable_noncash_deduction(Variable):
    """
    Line 12 on Montana itemized deductions schedule
    """

    value_type = float
    entity = TaxUnit
    label = "Montana charitable non-cash donation deduction"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT

    def formula(tax_unit, period, parameters):
        noncash_donations = tax_unit(period, ["charitable_non_cash_donations"])

        return noncash_donations



from fiscalsim_us.model_api import *


class sc_tuition_credit(Variable):
    """
    Tuition tax credit, refundable credit, line 21 on SC1040 2023
    Invidual Income Tax Return (see form I-319)
    """

    value_type = float
    entity = Person
    label = "South Carolina Tuition Credit"
    defined_for = "sc_tuition_credit_eligible"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.sc.gov/forms-site/Forms/I319_2023.pdf",
        "https://dor.sc.gov/forms-site/Forms/I319_2021.pdf#page=2",
        "https://www.scstatehouse.gov/code/t12c006.php",
        # South Carolina Legal Code | SECTION 12-6-3385 (A)
    )



from fiscalsim_us.model_api import *
from numpy import ceil


class me_dependent_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maine dependents exemption deduction"
    reference = "https://www.mainelegislature.org/legis/statutes/36/title36sec5219-SS.html"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.ME

    def formula(tax_unit, period, parameters):
        # First get total the number of dependents (line 1).
        num_dependents = tax_unit("tax_unit_dependents", period)

        # Get their Maine AGI (line 3).
        me_agi = tax_unit("me_agi", period)

        # Then get the Maine Dependents Exemptions part of the parameter tree.
        p = parameters(
            period
        ).gov.states.me.tax.income.deductions.dependent_exemption

        # Calculate the maximum dependents exemption amount (line 2).
        max_amount = num_dependents * p.amount

        # Get filing status.
        filing_status = tax_unit("filing_status", period)

        # Get the phaseout start amount based on filing status (line 4).
        phaseout_start = p.phaseout.start[filing_status]

        # Get the excess amount, if any, in thousands of dollars (rounded up) [lines 5 and 6].
        excess = max_(me_agi - phaseout_start, 0)
        increments = ceil(excess / p.phaseout.increment)

        # Calculate the excess part phase out amount (line 7).
        phase_out_amount = increments * p.phaseout.step

        # Return the max exemption amount minus the phaseout amount (line 8).
        return max_(max_amount - phase_out_amount, 0)



from fiscalsim_us.model_api import *


class taxable_self_employment_income(Variable):
    value_type = float
    entity = Person
    label = "Taxable self-employment income"
    definition_period = YEAR
    unit = USD
    reference = "https://www.law.cornell.edu/uscode/text/26/1402#a"

    def formula(person, period, parameters):
        irs = parameters(period).gov.irs
        se = irs.self_employment
        combined_rate = se.social_security_rate + se.medicare_rate
        deduction_rate = irs.ald.misc.employer_share * combined_rate
        net_se = person("sey", period) * (1 - deduction_rate)
        # Exclude net self-employment income below the reporting threshold.
        return where(net_se < se.net_earnings_exemption, 0, net_se)



from fiscalsim_us.model_api import *


class ar_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arkansas taxable income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/TaxBrackets_2022.pdf"
        "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2023_Final_AR1000ES.pdf"
    )
    defined_for = StateCode.AR



from fiscalsim_us.model_api import *


class sc_military_retirement_income_deduction_spouse(Variable):
    value_type = float
    entity = TaxUnit
    label = "South Carolina military retirement income deduction for spouse"
    defined_for = StateCode.SC
    unit = USD
    definition_period = YEAR



from fiscalsim_us.model_api import *


class employee_social_security_tax(Variable):
    value_type = float
    entity = Person
    label = "employee-side OASDI payroll tax"
    documentation = "Total liability for employee-side OASDI payroll tax."
    definition_period = YEAR
    unit = USD

    def formula(person, period, parameters):
        rate = parameters(period).gov.irs.payroll.social_security.rate.employee
        return rate * person("taxable_earnings_for_social_security", period)



from fiscalsim_us.model_api import *


class mt_casualty_loss_deduction(Variable):
    """
    Line 15 on Montana itemized deductions schedule
    """

    value_type = float
    entity = TaxUnit
    label = "Montana casualty and theft losses"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT

    def formula(tax_unit, period, parameters):
        deduction = (tax_unit, period, ["casualty_loss_deduction"])

        return deduction



from fiscalsim_us.model_api import *


class va_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "VA taxable income line 15 on form 760"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.VA

    def formula(tax_unit, period, parameters):
        line_14 = tax_unit("va_calc_line_14", period)

        line_9 = tax_unit("va_adj_gross_income", period)

        subtotal = line_9 - line_14

        return subtotal



from fiscalsim_us.model_api import *


class mn_medical_dental_deduction(Variable):
    """
    Minnesota medical and dental expense deduction
    Line 4 of 2022 M1SA, Minnesota Itemized Deductions
    """

    value_type = float
    entity = TaxUnit
    label = "MN medical and dental deductions"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MN

    def formula(tax_unit, period, parameters):
        expense = add(tax_unit, period, ["medical_expense"])
        fed_agi = tax_unit("adjusted_gross_income", period)
        mult = parameters(
            period
        ).gov.states.mn.tax.income.deductions.medical_and_dental_mult

        return max_(0, expense - mult * fed_agi)



from fiscalsim_us.model_api import *


class mi_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Michigan taxable income"
    defined_for = StateCode.MI
    unit = USD
    definition_period = YEAR



from fiscalsim_us.model_api import *


class mt_refundable_credits(Variable):
    """
    Line 22 on Montana individual tax return form 2
    """

    value_type = float
    entity = TaxUnit
    label = "Montana refundable credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT



from fiscalsim_us.model_api import *


class mn_unreimbursed_employee_expenses(Variable):
    """
    Line 20 of 2022 M1SA, Minnesota Itemized Deductions
    """

    value_type = float
    entity = TaxUnit
    label = "MN unreimbursed employee expenses"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MN



from pathlib import Path
from policyengine_core.taxbenefitsystems import TaxBenefitSystem
from fiscalsim_us.entities import *
from fiscalsim_us.parameters.gov.irs.uprating import (
    set_irs_uprating_parameter,
)
from policyengine_core.simulations import (
    Simulation as CoreSimulation,
    Microsimulation as CoreMicrosimulation,
    IndividualSim as CoreIndividualSim,
)
from fiscalsim_us.data import DATASETS, CPS_2023, EnhancedCPS_2023
from fiscalsim_us.variables.household.demographic.geographic.state.in_state import (
    create_50_state_variables,
)
from fiscalsim_us.tools.parameters import backdate_parameters
from fiscalsim_us.reforms import create_structural_reforms_from_parameters

from fiscalsim_us.reforms import create_structural_reforms_from_parameters

COUNTRY_DIR = Path(__file__).parent


class CountryTaxBenefitSystem(TaxBenefitSystem):
    parameters_dir = COUNTRY_DIR / "parameters"
    variables_dir = COUNTRY_DIR / "variables"
    auto_carry_over_input_variables = True
    basic_inputs = [
        "state_name",
        "employment_income",
        "age",
    ]
    modelled_policies = COUNTRY_DIR / "modelled_policies.yaml"

    def __init__(self):
        # We initialize our tax and benefit system with the general constructor
        super().__init__(entities)

        reform = create_structural_reforms_from_parameters(
            self.parameters, "2023-01-01"
        )
        if reform is not None:
            reform.apply(self)

        self.add_variables(*create_50_state_variables())

        self.parameters = set_irs_uprating_parameter(self.parameters)
        self.parameters = backdate_parameters(self.parameters)


system = CountryTaxBenefitSystem()


class Simulation(CoreSimulation):
    default_tax_benefit_system = CountryTaxBenefitSystem
    default_tax_benefit_system_instance = system
    default_role = "member"
    default_calculation_period = 2022
    default_input_period = 2022
    datasets = DATASETS

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        reform = create_structural_reforms_from_parameters(
            self.tax_benefit_system.parameters, "2023-01-01"
        )
        if reform is not None:
            self.apply_reform(reform)


class Microsimulation(CoreMicrosimulation):
    default_tax_benefit_system = CountryTaxBenefitSystem
    default_tax_benefit_system_instance = system
    default_dataset = CPS_2023
    default_dataset_year = 2023
    default_role = "member"
    default_calculation_period = 2023
    default_input_period = 2023
    datasets = DATASETS

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        reform = create_structural_reforms_from_parameters(
            self.tax_benefit_system.parameters, "2023-01-01"
        )
        if reform is not None:
            self.apply_reform(reform)


class IndividualSim(CoreIndividualSim):  # Deprecated
    tax_benefit_system = CountryTaxBenefitSystem
    entities = {entity.key: entity for entity in entities}
    default_dataset = CPS_2023

    default_roles = dict(
        tax_unit="member",
        spm_unit="member",
        household="member",
        family="member",
    )
    required_entities = [
        "tax_unit",
        "spm_unit",
        "household",
        "family",
    ]



from fiscalsim_us.model_api import *


class SSICategory(Enum):
    AGED = "Aged"
    BLIND = "Blind"
    DISABLED = "Disabled"
    NONE = "None"


class ssi_category(Variable):
    value_type = Enum
    entity = Person
    label = "SSI category"
    definition_period = YEAR
    possible_values = SSICategory
    default_value = SSICategory.NONE



from fiscalsim_us.model_api import *


class nm_exemptions(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Mexico income exemptions"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NM

    adds = [
        "nm_aged_blind_exemption",
        "nm_hundred_year_exemption",
        "nm_low_and_middle_income_exemption",
        "nm_medical_expense_exemption",
        "nm_social_security_income_exemption",
    ]



from fiscalsim_us.model_api import *


class ok_stc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Oklahoma sales tax credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/past-year/2021/511-Pkt-2021.pdf"
        "https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-Pkt.pdf"
    )
    defined_for = StateCode.OK

    def formula(tax_unit, period, parameters):
        # for details, see Form 538-S in the 511 packets referenced above
        p = parameters(period).gov.states.ok.tax.income.credits.sales_tax
        # determine TANF ineligibility
        tanf_ineligible = add(tax_unit, period, ["ok_tanf"]) > 0
        # determine income eligibility in two alternative ways
        income = tax_unit("ok_gross_income", period)
        # ... first way
        income_eligible1 = income <= p.income_limit1
        # ... second way
        count_dependents = tax_unit("tax_unit_dependents", period)
        has_dependents = count_dependents > 0
        elderly_head = tax_unit("age_head", period) >= p.age_minimum
        elderly_spouse = tax_unit("age_spouse", period) >= p.age_minimum
        has_elder = elderly_head | elderly_spouse
        disabled_head = tax_unit("head_is_disabled", period)
        disabled_spouse = tax_unit("spouse_is_disabled", period)
        has_disabled = disabled_head | disabled_spouse
        unit_eligible = has_dependents | has_elder | has_disabled
        income_eligible2 = unit_eligible & (income <= p.income_limit2)
        # determine overall eligibility
        eligible = ~tanf_ineligible & (income_eligible1 | income_eligible2)
        # calculate credit if eligible
        qualified_exemptions = tax_unit("num", period) + count_dependents
        return eligible * qualified_exemptions * p.amount



from fiscalsim_us.model_api import *


class uncapped_ssi(Variable):
    value_type = float
    entity = Person
    label = "Uncapped SSI"
    unit = USD
    documentation = "Maximum SSI, less countable income (can be below zero)."
    definition_period = YEAR

    def formula(person, period, parameters):
        amount = person("ssi_amount_if_eligible", period)
        meets_resource_test = person("meets_ssi_resource_test", period)
        eligible = person("is_ssi_eligible_individual", period)
        countable_income = person("ssi_countable_income", period)
        return (meets_resource_test & eligible) * (amount - countable_income)



from fiscalsim_us.model_api import *


class ssi_claim_is_joint(Variable):
    value_type = bool
    entity = Person
    label = "SSI claim is joint"
    definition_period = YEAR

    def formula(person, period, parameters):
        eligible = person("is_ssi_aged_blind_disabled", period)
        both_eligible = person.marital_unit.sum(eligible) == 2
        income_is_deemed = (
            person("ssi_income_deemed_from_ineligible_spouse", period) > 0
        )
        return both_eligible | income_is_deemed



from fiscalsim_us.model_api import *


class pre_c04600(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Personal exemption before phase-out"
    reference = "https://www.law.cornell.edu/uscode/text/26/151"
    unit = USD

    def formula(tax_unit, period, parameters):
        exemption = parameters(period).gov.irs.income.exemption
        return where(
            tax_unit("dsi", period),
            0,
            tax_unit("xtot", period) * exemption.amount,
        )



from fiscalsim_us.model_api import *
from policyengine_core.periods import instant


def get_irs_cpi(parameters: ParameterNode, year: int) -> float:
    cpi = parameters.gov.bls.cpi.c_cpi_u
    end = instant(f"{year}-08-01")
    start = end.offset(-MONTHS_IN_YEAR, MONTH)
    monthly_cpi_values = []
    for month in range(MONTHS_IN_YEAR):
        monthly_cpi_values += [cpi(start.offset(month, MONTH))]
    return sum(monthly_cpi_values) / MONTHS_IN_YEAR


def set_irs_uprating_parameter(parameters: ParameterNode) -> ParameterNode:
    uprating_index: Parameter = parameters.gov.irs.uprating
    for year in range(2010, 2030):
        irs_cpi = get_irs_cpi(parameters, year - 1)
        uprating_index.update(period=f"year:{year}-01-01:1", value=irs_cpi)
    return parameters



from fiscalsim_us.model_api import *
from random import randint, seed


class vehicles_owned(Variable):
    value_type = int
    entity = Person
    label = "Vehicles owned"
    definition_period = YEAR

    def formula(person, period, parameters):
        # We randomly split the household's vehicles between its adults
        household = person.household
        household_vehicles = household("household_vehicles_owned", period)
        is_adult = person("is_adult", period)
        num_adults_in_household = household.sum(is_adult)
        max_vehicles = household_vehicles.max()
        adult_rank = where(is_adult, household.members_position, 100)
        vehicles = is_adult * 0
        seed(0)
        for _ in range(int(max_vehicles)):
            # Pick a random adult in each household
            selected_adult = (
                randint(0, adult_rank[is_adult].max())
                % num_adults_in_household
            )
            maximum_reached = household.sum(vehicles) >= household_vehicles
            should_add_vehicle = ~maximum_reached & (
                adult_rank == selected_adult
            )
            vehicles += where(should_add_vehicle, 1, 0)
        return vehicles



from fiscalsim_us.model_api import *


class meets_school_meal_categorical_eligibility(Variable):
    value_type = bool
    entity = SPMUnit
    label = "School meal categorical eligibility"
    documentation = "Whether this SPM unit is eligible for free school meal via participation in other programs"
    definition_period = YEAR
    reference = "https://www.fns.usda.gov/cn/extending-categorical-eligibility-additional-children-household"

    def formula(spm_unit, period, parameters):
        programs = parameters(
            period
        ).gov.usda.school_meals.categorical_eligibility
        return np.any(
            [spm_unit(program, period) for program in programs], axis=0
        )



from fiscalsim_us.model_api import *


class sc_retirement_income_deduction_head(Variable):
    value_type = float
    entity = TaxUnit
    label = "South Carolina retirement income deduction for head"
    defined_for = StateCode.SC
    unit = USD
    definition_period = YEAR



import pandas as pd
from microdf import MicroDataFrame
from survey_enhance import Imputation
from typing import Tuple
from fiscalsim_us.system import system
from fiscalsim_us import Microsimulation
import numpy as np

PUF_FILE_PATH = "~/Downloads/puf_2015.csv"
PUF_DEMOGRAPHICS_FILE_PATH = "~/Downloads/demographics_2015.csv"

codebook = {
    "E00200": "employment_income",
    "E00300": "interest_received",
    "E00400": "tax_exempt_interest_income",
    "E00600": "dividends_included_in_agi",
    "E00650": "qualified_dividend_income",
    "E00700": "state_income_tax_refunds",
    "E00800": "alimony_received",
    "E00900": "self_employment_income",
    "E01000": "net_capital_gain_loss",
    "E01100": "capital_gain_distributions",
    "E01200": "other_gains_loss",
    "E01400": "taxable_ira_distribution",
    "E01500": "total_pensions_annuities_received",
    "E01700": "taxable_pension_income",
    "E02000": "schedule_e_net_income_loss",
    "E02100": "schedule_f_net_profit_loss",
    "E02300": "unemployment_compensation_in_agi",
    "E02400": "social_security",
    "E02500": "social_security_benefits_in_agi",
    "E03150": "total_deductible_ira_payments",
    "E03210": "student_loan_interest_deduction",
    "E03220": "educator_expenses",
    "E03230": "tuition_fees_deduction",
    "E03260": "self_employment_tax_deduction",
    "E03270": "self_employed_health_insurance_deduction",
    "E03240": "domestic_production_activities_deduction",
    "E03290": "health_savings_account_deduction",
    "E03300": "payments_to_keogh_accounts",
    "E03400": "forfeited_interest_penalty",
    "E03500": "alimony_paid",
    "E00100": "adjusted_gross_income",
    "P04470": "total_deductions_standard_itemized",
    "E04600": "exemption_amount",
    "E04800": "taxable_income",
    "E05100": "tax_on_taxable_income",
    "E05200": "computed_regular_tax",
    "E05800": "income_tax_before_credits",
    "E06000": "income_subject_to_tax",
    "E06200": "marginal_tax_base",
    "E06300": "tax_generated_tax_rate_tables",
    "E09600": "alternative_minimum_tax",
    "E07180": "child_dependent_care_credit",
    "E07200": "elderly_disabled_credit",
    "E07220": "child_tax_credit",
    "E07230": "education_credits",
    "E07240": "retirement_savings_credit",
    "E07260": "residential_energy_credit",
    "E07300": "foreign_tax_credit",
    "E07400": "general_business_credit",
    "E07600": "credit_prior_year_minimum_tax",
    "P08000": "other_tax_credits",
    "E07150": "total_tax_credit_soi",
    "E06500": "total_income_tax",
    "E08800": "income_tax_after_credits_soi",
    "E09400": "self_employment_tax",
    "E09700": "recapture_taxes",
    "E09730": "total_additional_medicare_tax",
    "E09740": "net_investment_income_tax",
    "E09750": "health_care_individual_responsibility_payment",
    "E09800": "social_security_tax_tip_income",
    "E09900": "penalty_tax_ira",
    "E10300": "total_tax_liability_soi",
    "E10700": "income_tax_withheld",
    "E10900": "estimated_tax_payments",
    "E10960": "refundable_american_opportunity_credit",
    "E59560": "earned_income_eic",
    "E59680": "eic_offset_income_tax_before_credits",
    "E59700": "eic_offset_other_taxes_except_advance_eic",
    "E59720": "eic_refundable_portion",
    "E11550": "refundable_prior_year_minimum_tax_credit",
    "E11560": "net_premium_tax_credit",
    "E11561": "net_premium_tax_credit_offset_income_tax_before_credits",
    "E11562": "net_premium_tax_credit_offset_other_taxes",
    "E11563": "net_premium_tax_credit_refundable_portion",
    "E11070": "additional_child_tax_credit",
    "E11100": "amount_paid_form_4868_request_extension",
    "E11200": "excess_fica_rrta",
    "E11300": "credit_federal_tax_special_fuels_oils",
    "E11400": "regulated_investment_company_credit",
    "E11601": "total_refundable_credits_offset_income_tax_before_credits",
    "E11602": "total_refundable_credits_offset_other_taxes",
    "E11603": "total_refundable_credits_refundable_parts",
    "E10605": "total_tax_payments_soi",
    "E11900": "balance_due_overpayment",
    "E12000": "credit_elect",
    "E12200": "predetermined_estimated_tax_penalty",
    "E17500": "medical_dental_expenses_reduction_agi_limit",
    "E18400": "state_local_taxes",
    "E18500": "real_estate_tax_deductions",
    "E19200": "total_interest_paid_deduction",
    "E19550": "qualified_mortgage_insurance_premiums",
    "E19800": "cash_contributions",
    "E20100": "other_than_cash_contributions",
    "E19700": "contributions_deduction_total",
    "E20550": "unreimbursed_employee_business_expense",
    "E20600": "tax_preparation_fee",
    "E20400": "miscellaneous_deductions_agi_limitation_total",
    "E20800": "net_limited_miscellaneous_deductions",
    "E20500": "net_casualty_theft_loss",
    "E21040": "itemized_deduction_limitation",
    "P22250": "short_term_capital_gains",
    "E22320": "long_term_capital_gains_1",
    "E22370": "schedule_d_capital_gain_distributions",
    "P23250": "long_term_capital_gains",
    "E24515": "unrecaptured_section_1250_gain",
    "E24516": "capital_gain_less_investment_expense",
    "E24518": "28_percent_rate_gain_loss",
    "E24560": "non_schedule_d_tax",
    "E24598": "schedule_d_15_percent_tax_amount",
    "E24615": "schedule_d_25_percent_tax_amount",
    "E24570": "schedule_d_28_percent_tax_amount",
    "P25350": "rental_income",
    "P25380": "rent_royalty_interest_expenses",
    "E25550": "total_depreciation_depletion_all_property",
    "P25700": "rent_royalty_net_income_loss",
    "E25820": "deductible_rental_loss",
    "E25850": "rent_royalty_net_income",
    "E25860": "rent_royalty_net_loss",
    "E25940": "total_passive_income_partnerships",
    "E25980": "total_non_passive_income_partnerships",
    "E25920": "total_passive_loss_partnerships",
    "E25960": "total_non_passive_loss_partnerships",
    "E26110": "partnership_section_179_expense_deduction",
    "E26170": "total_passive_income_small_business_corp",
    "E26190": "total_non_passive_income_small_business_corp",
    "E26160": "total_passive_loss_small_business_corp",
    "E26180": "total_non_passive_loss_small_business_corp",
    "E26270": "partnership_s_corp_income",
    "E26100": "s_corp_section_179_expense_deduction",
    "E26390": "total_income_estate_trust",
    "E26400": "total_loss_estate_trust",
    "E27200": "farm_rent_income",
    "E30400": "self_employment_income_ss_tax_primary",
    "E30500": "self_employment_income_ss_tax_secondary",
    "E32800": "qualifying_individuals_expenses_form_2441",
    "E33000": "expenses_limited_to_earned_income_form_2441",
    "E53240": "work_opportunity_jobs_general_business_credit",
    "E53280": "research_experimentation_general_business_credit",
    "E53300": "low_income_housing_credit",
    "E53317": "employer_credit_social_security_tax_tips",
    "E58950": "total_investment_interest_expense_form_4952",
    "E58990": "investment_income_elected_amount_form_4952",
    "P60100": "net_operating_loss_tax_preference_adjustments_form_6251",
    "P61850": "total_adjustments_preferences_form_6251",
    "E60000": "taxable_income_amt_form_6251",
    "E62100": "alternative_minimum_taxable_income",
    "E62900": "alternative_tax_foreign_tax_credit",
    "E62720": "alternative_minimum_schedule_d_less_investment_interest",
    "E62730": "alternative_minimum_schedule_d_unrecaptured_section_1250_gain",
    "E62740": "alternative_minimum_capital_gain_amount",
    "P65300": "total_passive_net_income_form_8582",
    "P65400": "total_passive_losses_form_8582",
    "E68000": "total_losses_allowed_passive_activities",
    "E82200": "carry_forward_minimum_tax_credit_form_8801",
    "T27800": "farm_income",
    "S27860": "tentative_current_prior_year_tax_schedule_j",
    "P27895": "actual_prior_year_tax_schedule_j",
    "P87482": "american_opportunity_qualified_expenses_form_8863",
    "E87521": "american_opportunity_credit",
    "E87530": "lifetime_learning_total_qualified_expenses",
    "E87550": "lifetime_learning_credit",
    "P86421": "bond_purchase_amount_form_8888",
    "E85050": "total_rental_real_estate_royalties_partnerships_s_corps_trusts",
    "E85090": "total_net_gain_loss_disposition_property",
    "E85120": "total_investment_income",
    "E85180": "total_deductions_modifications",
    "E85570": "dependents_modified_adjusted_gross_income_amount_form_8962",
    "E85595": "annual_contribution_health_care_amount",
    "E85600": "monthly_contribution_health_care_amount",
    "E85770": "total_premium_tax_credit_amount",
    "E85775": "advance_premium_tax_credit_amount",
    "E85785": "excess_advance_payment_premium_tax_credit",
    "E85790": "repayment_limitation_amount",
    "RECID": "return_id",
    "S006": "decimal_weight",
    "S008": "sample_count",
    "S009": "population_count",
    "WSAMP": "sample_code",
    "TXRT": "marginal_tax_rate",
}

codebook.update(
    {
        "AGIR1": "adjusted_gross_income_band",
        "CLAIM8965": "health_coverage_exemptions",
        "DSI": "dependent_status_indicator",
        "EFI": "electronic_filing_indicator",
        "EIC": "earned_income_credit_code",
        "ELECT": "presidential_election_campaign_fund_boxes",
        "FDED": "form_of_deduction_code",
        "FLPDYR": "filing_accounting_period_year",
        "FLPDMO": "filing_accounting_period_month",
        "F2441": "form_2441_child_care_credit_qualified_individual",
        "F3800": "form_3800_general_business_credit",
        "F6251": "form_6251_alternative_minimum_tax",
        "F8582": "form_8582_passive_activity_loss_limitation",
        "F8606": "form_8606_nondeductible_ira_contributions",
        "F8829": "form_8829_expenses_home_business_use",
        "F8867": "form_8867_paid_preparer_earned_income_credit_checklist",
        "F8949": "form_8949_sales_dispositions_capital_assets",
        "F8959": "form_8959_additional_medicare_tax",
        "F8960": "form_8960_net_investment_income_tax",
        "F8962": "form_8962_premium_tax_credit",
        "F8965": "form_8965_health_coverage_exemptions",
        "IE": "itemized_deductions_election_indicator",
        "MARS": "marital_filing_status",
        "MIDR": "married_filing_separately_itemized_deductions_requirement_indicator",
        "N24": "number_children_child_tax_credit",
        "N25": "number_qualified_students_lifetime_learning_credit",
        "N30": "number_qualified_students_american_opportunity_credit",
        "PREP": "tax_preparer",
        "PREMNTHS": "months_enrolled_health_insurance_marketplace",
        "SCHB": "schedule_b_indicator",
        "SCHCF": "schedule_c_or_f_indicator",
        "SCHE": "schedule_e_indicator",
        "TFORM": "form_of_return",
        "TXST": "tax_status",
        "XFPT": "primary_taxpayer_exemption",
        "XFST": "secondary_taxpayer_exemption",
        "XOCAH": "exemptions_children_living_at_home",
        "XOCAWH": "exemptions_children_living_away_from_home",
        "XOODEP": "exemptions_other_dependents",
        "XOPAR": "exemptions_parents_living_at_away_from_home",
        "XTOT": "total_exemptions",
        "XTOT8962": "number_exemptions_form_8962",
        "XTOT8965": "number_exemptions_form_8965",
    }
)

codebook.update(
    {
        "AGEDP1": "age_dependent_1",
        "AGEDP2": "age_dependent_2",
        "AGEDP3": "age_dependent_3",
        "AGERANGE": "age_range_primary_filer",
        "EARNSPLIT": "earnings_split_joint_returns",
        "GENDER": "gender_primary_filer",
        "RECID": "return_id",
    }
)

CPS_DATASET = "cps_2023"


DEMOGRAPHIC_VARIABLES = [
    "age_dependent_1",
    "age_dependent_2",
    "age_dependent_3",
    "age_range_primary_filer",
    "earnings_split_joint_returns",
    "gender_primary_filer",
]

FINANCIAL_VARIABLES = [
    column
    for column in codebook.values()
    if column
    not in DEMOGRAPHIC_VARIABLES
    + [
        "return_id",
        "decimal_weight",
        "sample_count",
        "population_count",
        "sample_code",
    ]
]

FINANCIAL_SUBSET = [
    "employment_income",
    "self_employment_income",
    "partnership_s_corp_income",
    "farm_income",
    "farm_rent_income",
    "short_term_capital_gains",
    "long_term_capital_gains",
    "taxable_interest_income",
    "tax_exempt_interest_income",
    "rental_income",
    "qualified_dividend_income",
    "non_qualified_dividend_income",
    "taxable_pension_income",
    "social_security",
]


def load_puf() -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Load the PUF and demographics data.

    Returns:
        puf (pd.DataFrame): The PUF data.
        puf_with_demographics (pd.DataFrame): The subset of the PUF data that has demographic information.
    """
    puf = pd.read_csv("~/Downloads/puf_2015.csv")
    demographics = pd.read_csv("~/Downloads/demographics_2015.csv")

    puf = puf.dropna()
    demographics = demographics.dropna()

    puf.S006 = puf.S006 / 100

    puf["taxable_interest_income"] = (
        puf.E00300 - puf.E00400
    )  # Taxable interest income = interest received - tax-exempt interest income
    puf["non_qualified_dividend_income"] = (
        puf.E00600 - puf.E00650
    )  # Non-qualified dividend income = dividends included in AGI - qualified dividend income

    puf = puf.rename(columns=codebook)
    demographics = demographics.rename(columns=codebook)
    return puf, demographics


def impute_missing_demographics(
    puf: pd.DataFrame,
    demographics: pd.DataFrame,
) -> pd.DataFrame:
    """Impute missing demographic information from the PUF.

    Args:
        puf (pd.DataFrame): The PUF data.
        puf_with_demographics (pd.DataFrame): The subset of the PUF data that has demographic information.

    Returns:
        puf_with_imputed_demographics (pd.DataFrame): The PUF data with imputed demographic information.
    """

    demographics_from_puf = Imputation()
    puf_with_demographics = puf[
        puf.return_id.isin(demographics.return_id)
    ].merge(demographics, on="return_id")

    soi = system.parameters.calibration.gov.irs.soi

    # Uprate the financial subset

    for variable_name in FINANCIAL_SUBSET:
        value_in_2015 = soi.children[variable_name]("2015-01-01")
        value_in_2023 = soi.children[variable_name]("2023-01-01")
        uprating_factor = value_in_2023 / value_in_2015
        puf_with_demographics[variable_name] = (
            puf_with_demographics[variable_name] * uprating_factor
        )

    demographics_from_puf.train(
        puf_with_demographics[FINANCIAL_VARIABLES],
        puf_with_demographics[DEMOGRAPHIC_VARIABLES],
    )

    puf_without_demographics = puf[
        ~puf.return_id.isin(puf_with_demographics.return_id)
    ].reset_index()
    puf_without_demographics.marital_filing_status = (
        puf_without_demographics.marital_filing_status.replace(
            {
                0: 1,
            }
        )
    )  # Aggregated returns -> single
    predicted_demographics = demographics_from_puf.predict(
        puf_without_demographics[FINANCIAL_VARIABLES]
    )
    puf_with_imputed_demographics = pd.concat(
        [puf_without_demographics, predicted_demographics], axis=1
    )

    weighted_puf_with_demographics = MicroDataFrame(
        puf_with_demographics, weights="decimal_weight"
    )
    weighted_puf_with_imputed_demographics = MicroDataFrame(
        puf_with_imputed_demographics, weights="decimal_weight"
    )

    puf_combined = pd.concat(
        [
            weighted_puf_with_demographics,
            weighted_puf_with_imputed_demographics,
        ]
    )
    puf_combined = MicroDataFrame(puf_combined, weights="decimal_weight")

    return puf_combined


def generate_puf_style_cps() -> pd.DataFrame:
    """Generate a PUF-style table from the CPS.

    Returns:
        cps (pd.DataFrame): The CPS data.
    """

    sim = Microsimulation(dataset=CPS_DATASET)

    cps_demographics = pd.DataFrame(index=sim.calculate("tax_unit_id").values)

    df = sim.calculate_dataframe(
        ["age", "tax_unit_id", "is_tax_unit_dependent"]
    )
    df = df[df.is_tax_unit_dependent]
    df_sorted = df.sort_values(["tax_unit_id", "age"])
    df_sorted["rank"] = df_sorted.groupby("tax_unit_id")["age"].rank()

    df_sorted["age_dependent_1"] = np.where(
        df_sorted["rank"] == 1, df_sorted["age"], -1
    )
    df_sorted["age_dependent_2"] = np.where(
        df_sorted["rank"] == 2, df_sorted["age"], -1
    )
    df_sorted["age_dependent_3"] = np.where(
        df_sorted["rank"] == 3, df_sorted["age"], -1
    )

    df_sorted_maxed = df_sorted.groupby("tax_unit_id").max()

    cps_demographics["age_dependent_1"] = df_sorted_maxed["age_dependent_1"]
    cps_demographics["age_dependent_2"] = df_sorted_maxed["age_dependent_2"]
    cps_demographics["age_dependent_3"] = df_sorted_maxed["age_dependent_3"]

    cps_demographics = cps_demographics.fillna(-1)

    # Define the age bins and labels
    bins = [-np.inf, -1, 4, 12, 16, 18, 23, np.inf]
    labels = [0, 1, 2, 3, 4, 5, 6]

    # Create AGEDP1, AGEDP2, AGEDP3 based on the categories
    for col in ["age_dependent_1", "age_dependent_2", "age_dependent_3"]:
        cps_demographics[col] = pd.cut(
            cps_demographics[col], bins=bins, labels=labels, right=True
        )

    cps_demographics.reset_index(inplace=True)
    cps_demographics = cps_demographics[
        ["age_dependent_1", "age_dependent_2", "age_dependent_3"]
    ]

    cps_demographics["age_range_primary_filer"] = sim.calculate(
        "age_head"
    ).values

    bins_head = [-np.inf, -1, 25, 34, 44, 54, 64, np.inf]
    labels_head = [0, 1, 2, 3, 4, 5, 6]

    cps_demographics["age_range_primary_filer"] = pd.cut(
        cps_demographics["age_range_primary_filer"],
        bins=bins_head,
        labels=labels_head,
        right=True,
    )

    is_male = sim.calculate("is_male")
    is_head = sim.calculate("is_tax_unit_head")
    male_head = sim.map_result(is_male * is_head, "person", "tax_unit")
    tax_unit_filer_gender = np.where(male_head, 1, 2)

    cps_demographics["gender_primary_filer"] = tax_unit_filer_gender

    filer_earned = sim.calculate("filer_earned")
    spouse_earned = sim.calculate("spouse_earned")
    filing_status = sim.calculate("filing_status")

    def determine_earning_split_value(
        filer: float, spouse: float, filing_status: str
    ) -> int:
        if filing_status != "JOINT":
            return 0
        if filer + spouse <= 0:
            return 1
        ratio_filer = filer / (filer + spouse)
        if ratio_filer >= 0.75:
            return 1
        elif ratio_filer >= 0.25:
            return 2
        else:
            return 3

    cps_demographics["earnings_split_joint_returns"] = np.vectorize(
        determine_earning_split_value
    )(filer_earned, spouse_earned, filing_status)
    cps_demographics["tax_unit_weight"] = sim.calculate(
        "tax_unit_weight"
    ).values
    return cps_demographics


def impute_puf_financials_to_cps(
    cps_demographics: pd.DataFrame,
    puf: pd.DataFrame,
):
    """Impute PUF financials to the CPS.

    Args:
        cps_demographics (pd.DataFrame): The CPS data with demographic information.
        puf (pd.DataFrame): The PUF data.

    Returns:
        cps_imputed (pd.DataFrame): The CPS data with imputed financial information.
    """
    income_from_demographics = Imputation()

    income_from_demographics.train(
        puf[DEMOGRAPHIC_VARIABLES],
        puf[FINANCIAL_SUBSET],
        sample_weight=puf.decimal_weight,
    )

    cps_financial_predictions = income_from_demographics.predict(
        cps_demographics[DEMOGRAPHIC_VARIABLES],
        mean_quantile=[0.9] + [0.5] * (len(FINANCIAL_SUBSET) - 1),
    )
    cps_imputed = pd.concat(
        [cps_demographics, cps_financial_predictions], axis=1
    )
    cps_imputed = MicroDataFrame(
        cps_imputed, weights=cps_demographics.tax_unit_weight
    )

    return cps_imputed


def project_tax_unit_cps_to_person_level(
    puf_style_cps: pd.DataFrame,
) -> pd.DataFrame:
    """Project tax unit CPS to person level.

    Args:
        puf_style_cps (pd.DataFrame): The CPS data with imputed financial information.

    Returns:
        person_df (pd.DataFrame): The CPS data with imputed financial information projected to the person level.
    """
    sim = Microsimulation(dataset=CPS_DATASET)
    person_df = pd.DataFrame(
        dict(
            person_id=sim.calculate("person_id").values,
            tax_unit_id=sim.calculate("person_tax_unit_id").values,
        )
    )

    person_is_tax_filer_head = sim.calculate("is_tax_unit_head").values

    for variable in FINANCIAL_SUBSET:
        cps_original_value = sim.calculate(variable).values
        cps_tax_unit_original_total = sim.map_result(
            sim.map_result(cps_original_value, "person", "tax_unit"),
            "tax_unit",
            "person",
        )
        cps_share_of_tax_unit_original_total = (
            cps_original_value / cps_tax_unit_original_total
        )
        cps_share_of_tax_unit_original_total = np.where(
            np.isnan(cps_share_of_tax_unit_original_total),
            person_is_tax_filer_head,
            cps_share_of_tax_unit_original_total,
        )
        mapped_down_imputed_values = sim.map_result(
            puf_style_cps[variable].values, "tax_unit", "person"
        )
        person_df[variable] = (
            cps_share_of_tax_unit_original_total * mapped_down_imputed_values
        )

    person_df = person_df.fillna(0)
    return person_df


def puf_imputed_cps_person_level(
    verbose: bool = False,
) -> pd.DataFrame:
    """Generate a PUF-imputed CPS at the person level.

    Args:
        verbose (bool): Whether to print progress statements.

    Returns:
        person_level_puf_imputed_cps (pd.DataFrame): The PUF-imputed CPS at the person level.
    """
    if verbose:
        print("Loading PUF and demographics")
    puf, demographics = load_puf()

    if verbose:
        print("Imputing missing demographics")
    puf = impute_missing_demographics(puf, demographics)

    if verbose:
        print("Generating PUF-style CPS")
    puf_style_cps = generate_puf_style_cps()

    if verbose:
        print("Imputing PUF financials to CPS")
    puf_imputed_cps = impute_puf_financials_to_cps(puf_style_cps, puf)

    if verbose:
        print("Projecting tax unit CPS to person level")
    person_level_puf_imputed_cps = project_tax_unit_cps_to_person_level(
        puf_imputed_cps
    )

    if verbose:
        print("Done")
    return person_level_puf_imputed_cps


if __name__ == "__main__":
    puf_imputed_cps_person_level(verbose=True)



from fiscalsim_us.model_api import *
from fiscalsim_us.variables.gov.ssa.ssi.eligibility.income._apply_ssi_exclusions import (
    _apply_ssi_exclusions,
)


class ssi_countable_income(Variable):
    value_type = float
    entity = Person
    label = "SSI countable income"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/42/1382a#b"

    def formula(person, period, parameters):
        personal_earned_income = person("ssi_earned_income", period)
        personal_unearned_income = person("ssi_unearned_income", period)
        parental_income_deemed_as_unearned_income = person(
            "ssi_unearned_income_deemed_from_ineligible_parent", period
        )

        personal_income = _apply_ssi_exclusions(
            personal_earned_income,
            personal_unearned_income
            + parental_income_deemed_as_unearned_income,
            parameters,
            period,
        )

        income_from_spouse = person(
            "ssi_income_deemed_from_ineligible_spouse", period
        )

        # Ensure the spouse whose income is deemed to their spouse loses their income.
        has_donated_income = (
            person.marital_unit.sum(income_from_spouse) > income_from_spouse
        )

        is_ssi_claimant = person("is_ssi_eligible_individual", period)

        return where(
            has_donated_income | ~is_ssi_claimant,
            0,
            personal_income + income_from_spouse,
        )



from fiscalsim_us.model_api import *


class pell_grant(Variable):
    value_type = float
    entity = Person
    label = "Pell Grant"
    definition_period = YEAR
    unit = USD

    def formula(person, period, parameters):
        coa = person("cost_of_attending_college", period)
        months_in_school = person("pell_grant_months_in_school", period)
        efc = person("pell_grant_efc", period)
        p = parameters(period).gov.ed.pell_grant
        unbounded = coa - efc
        capped = min_(unbounded, p.amount.max)
        amount = where(capped < p.amount.min, 0, capped)
        return amount * (months_in_school / p.months_in_school_year)



from fiscalsim_us.model_api import *


class ia_pension_exclusion(Variable):
    value_type = float
    entity = Person
    label = "Iowa pension exclusion"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.iowa.gov/sites/default/files/2023-01/2021%20Expanded%20Instructions_010323.pdf#page=26"
        "https://tax.iowa.gov/sites/default/files/2023-03/2022%20Expanded%20Instructions_022023.pdf#page=26"
    )
    defined_for = StateCode.IA

    def formula(person, period, parameters):
        p = parameters(period).gov.states.ia.tax.income.pension_exclusion
        # determine eligibility for pension exclusion
        is_head = person("is_tax_unit_head", period)
        is_spouse = person("is_tax_unit_spouse", period)
        # ... determine age eligibility
        an_elder = person("age", period) >= p.minimum_age
        is_elderly = (is_head & an_elder) | (is_spouse & an_elder)
        # ... determine disability eligiblity
        has_disability = person("is_permanently_and_totally_disabled", period)
        is_disabled = (is_head & has_disability) | (is_spouse & has_disability)
        # ... determine exclusion eligibility
        is_eligible = is_elderly | is_disabled
        # determine pension exclusion amount
        pension = person("taxable_pension_income", period)
        filing_status = person.tax_unit("filing_status", period)
        exclusion_cap = p.maximum_amount[filing_status]
        return is_eligible * min_(pension, exclusion_cap)



from fiscalsim_us.model_api import *
from policyengine_core.taxscales import MarginalRateTaxScale


class ny_supplemental_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "NY supplemental income tax"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NY

    def formula(tax_unit, period, parameters):
        ny_taxable_income = tax_unit("ny_taxable_income", period)
        ny_agi = tax_unit("ny_agi", period)
        tax = parameters(period).gov.states.ny.tax.income
        sup_tax = tax.supplemental

        filing_status = tax_unit("filing_status", period)
        status = filing_status.possible_values
        statuses = [
            status.SINGLE,
            status.JOINT,
            status.HEAD_OF_HOUSEHOLD,
            status.SURVIVING_SPOUSE,
            status.SEPARATE,
        ]
        in_each_status = [filing_status == s for s in statuses]

        rates = tax.main
        single = rates.single
        joint = rates.joint
        hoh = rates.head_of_household
        widow = rates.widow
        separate = rates.separate
        scales = [single, joint, hoh, widow, separate]

        previous_agi_threshold = select(
            in_each_status,
            [
                get_previous_threshold(ny_agi, scale.thresholds)
                for scale in scales
            ],
        )

        applicable_amount = max_(
            0, ny_agi - max_(previous_agi_threshold, sup_tax.min_agi)
        )

        phase_in_fraction = min_(
            1,
            applicable_amount / sup_tax.phase_in_length,
        )

        rate = select(
            in_each_status,
            [
                scale.marginal_rates(
                    max_(sup_tax.min_agi + 1, ny_taxable_income)
                )
                for scale in scales
            ],
        )

        target_tax = rate * ny_taxable_income
        difference = target_tax - tax_unit("ny_main_income_tax", period)
        return phase_in_fraction * difference



from fiscalsim_us.model_api import *
from policyengine_core.parameters.operations import (
    homogenize_parameter_structures,
)
from policyengine_core.simulations import Simulation


class second_lowest_silver_plan_cost(Variable):
    value_type = float
    entity = TaxUnit
    label = "Second-lowest silver plan cost"
    unit = USD
    definition_period = YEAR
    defined_for = "is_ptc_eligible"
    hidden_input = True

    def formula(tax_unit, period, parameters):
        simulation: Simulation = tax_unit.simulation
        if (
            simulation.get_holder("reported_slspc").get_array(period)
            is not None
        ):
            # If the user has provided a value for the second-lowest silver plan
            # cost, use that.
            return simulation.calculate("reported_slspc", period)
        parameter_tree = tax_unit.simulation.tax_benefit_system.parameters
        if not hasattr(parameter_tree.gov.hhs.medicaid, "geography"):
            medicaid_parameters = ParameterNode(
                directory_path=REPO
                / "data"
                / "parameters"
                / "gov"
                / "hhs"
                / "medicaid"
                / "geography"
            )
            medicaid_parameters = homogenize_parameter_structures(
                medicaid_parameters,
                tax_unit.simulation.tax_benefit_system.variables,
            )
            parameter_tree.gov.hhs.medicaid.add_child(
                "geography", medicaid_parameters
            )
        person = tax_unit.members
        household = person.household
        area = household("medicaid_rating_area", period)
        state = household("state_code_str", period)
        age = person("age", period)
        slspc = parameter_tree.gov.hhs.medicaid.geography.second_lowest_silver_plan_cost(
            period
        )
        age_code = select(
            [
                age < 21,
                age < 64,
                age >= 64,
            ],
            [
                "0-20",
                age.astype(int).astype(str),
                "64+",
            ],
        )
        eligible = person.tax_unit("is_ptc_eligible", period)
        per_person_cost = index_(
            into=slspc,
            indices=[state, area, age_code],
            where=eligible,
        )
        return tax_unit.sum(per_person_cost)



from fiscalsim_us.model_api import *


class mt_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Montana EITC"
    unit = USD
    definition_period = YEAR
    reference = "https://leg.mt.gov/bills/mca/title_0150/chapter_0300/part_0230/section_0180/0150-0300-0230-0180.html"

    def formula(tax_unit, period, parameters):
        eitc = tax_unit("earned_income_tax_credit", period)
        filing_status = tax_unit("filing_status", period)
        rate = parameters(period).gov.states.mt.tax.income.credits.eitc.rate
        return where(
            filing_status == filing_status.possible_values.SEPARATE,
            0,
            eitc * rate,
        )



from fiscalsim_us.model_api import *


class ut_refundable_credits(Variable):
    """
    Line 38 on Utah 2022 Individual Income Tax return form TC-40. These
    credits include the following categories and are listed at
    https://incometax.utah.gov/credits with their accompanying codes.
    * 39: Renewable commercial energy systems credit
    * 41: Special needs adoption credit
    * 47: Agricultural off-highway gas/undyed diesel credit
    * 48: Farm operation hand tools credit
    """

    value_type = float
    entity = TaxUnit
    label = "UT refundable credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.UT



from fiscalsim_us.model_api import *


class mt_home_mortage_interest_deduction(Variable):
    """
    Line 9 on itemized deductions schedule
    Home mortgage interst and points
    """

    value_type = float
    entity = TaxUnit
    label = "Montana home mortage interest deduction"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT



from fiscalsim_us.model_api import *


class mt_tax_after_nonrefundable_credits(Variable):
    """
    Line 20 on Montana individual tax return form 2
    """

    value_type = float
    entity = TaxUnit
    label = "Montana refundable credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT

    def formula(tax_unit, period, parameter):
        nonrefundable = tax_unit("mt_nonrefundable_credits")
        incometaxbeforecredits = tax_unit(
            "mt_income_tax_before_refundable_credits"
        )
        return incometaxbeforecredits - nonrefundable



from fiscalsim_us.model_api import *


class mn_taxes_paid_deducion(Variable):
    """
    Line 10 of 2022 M1SA, Minnesota Itemized Deductions
    """

    value_type = float
    entity = TaxUnit
    label = "MN taxes paid deduction"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MN

    def formula(tax_unit, period, parameters):
        real_estate = add(tax_unit, period, ["real_estate_taxes"])
        personal_property = tax_unit("mn_personal_property_tax", period)
        filing_status = tax_unit("filing_status", period)

        ceiling = parameters(
            period
        ).gov.states.mn.tax.income.deductions.taxes_paid_ceiling[filing_status]

        return min_(ceiling, real_estate + personal_property)



from fiscalsim_us.model_api import *


class ecpa_filer_credit(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    unit = USD
    label = "End Child Poverty Act Filer Credit"
    reference = "https://tlaib.house.gov/sites/tlaib.house.gov/files/EndChildPovertyAct.pdf"

    def formula_2022(tax_unit, period, parameters):
        # Filer credit.
        # Define eligibility based on age.
        age_head = tax_unit("age_head", period)
        age_spouse = tax_unit("age_spouse", period)
        p = parameters(
            period
        ).gov.contrib.congress.tlaib.end_child_poverty_act.filer_credit

        head_qualifies = (age_head >= p.eligibility.min_age) & (
            age_head <= p.eligibility.max_age
        )
        spouse_qualifies = (age_spouse >= p.eligibility.min_age) & (
            age_spouse <= p.eligibility.max_age
        )
        filer_credit_eligible = head_qualifies | spouse_qualifies
        # Get maximum amount.
        filing_status = tax_unit("filing_status", period)
        max_filer_credit = p.amount[filing_status]
        # Phase out.
        agi = tax_unit("adjusted_gross_income", period)
        phase_out_start = p.phase_out.start[filing_status]
        excess = max_(agi - phase_out_start, 0)
        reduction = excess * p.phase_out.rate
        # Compute final amount.
        return filer_credit_eligible * max_(max_filer_credit - reduction, 0)



from fiscalsim_us.model_api import *


class aged_head(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Aged and or blind head and spouse count"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/63#f"

    def formula(tax_unit, period, parameters):
        age_threshold = parameters(
            period
        ).gov.irs.deductions.standard.aged_or_blind.age_threshold
        return tax_unit("age_head", period) >= age_threshold



from fiscalsim_us.model_api import *


class ut_additions_to_income(Variable):
    """
    Line 5 on Utah 2022 Individual Income Tax return form TC-40. These
    additions to income include the following categories and are listed at
    https://incometax.utah.gov/additions with their accompanying codes.
    * 51: Lump sum distribution
    * 53: Medical care savings account (MSA) addback
    * 54: my529 addback
    * 56: Child's income excluded from parent's return
    * 57: Municipal bond interest
    * 60: Untaxed income of a resident trust
    * 61: Untaxed income of a nonresident trust
    * 67: Tax paid on behalf of a pass-through entity taxpayer
    * 68: Payroll Protection Program grant or loan addback
    * 69: Equitable adjustments
    """

    value_type = float
    entity = TaxUnit
    label = "UT additions to income"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.UT



from fiscalsim_us.model_api import *


class ut_income_tax_before_refundable_credits(Variable):
    """
    Line 32 of Utah 2022 Individual Income Tax return form TC-40.
    """

    value_type = float
    entity = TaxUnit
    label = "UT income tax before refundable credits"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        income_tax_after_credits = tax_unit(
            "ut_income_tax_after_credits", period
        )
        apportionable_nonrefundable_credits = tax_unit(
            "ut_apportionable_nonrefundable_credits", period
        )
        nonapportionable_nonrefundable_credits = tax_unit(
            "ut_nonapportionable_nonrefundable_credits", period
        )
        income_tax_before_additions = max_(
            income_tax_after_credits
            - apportionable_nonrefundable_credits
            - nonapportionable_nonrefundable_credits,
            0,
        )
        recapture_low_inc_housing_credit = tax_unit(
            "ut_recapture_low_inc_housing_credit", period
        )
        return income_tax_before_additions + recapture_low_inc_housing_credit



from fiscalsim_us.model_api import *


class va_age_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "VA age deduction https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2021-760-instructions.pdf for information"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.VA

    def formula(tax_unit, period, parameters):
        age_of_head = tax_unit("age_head", period)
        age_of_spouse = tax_unit("age_spouse", period)
        filing_status = tax_unit("filing_status", period)
        filing_statuses = filing_status.possible_values
        federal_agi = tax_unit("adjusted_gross_income", period)
        spouse_agi = tax_unit("spouse_separate_adjusted_gross_income", period)
        you_fdca = tax_unit("va_fixed_date_conformity_additions", period)
        spouse_fdca = 0  # change this
        you_fdcs = tax_unit("va_fixed_date_conformity_subtractions", period)
        spouse_fdcs = 0  # change this

        age_deduction_count = where(
            (age_of_head > 65)
            & (
                (filing_status == filing_statuses.SINGLE)
                | (filing_status == filing_statuses.HEAD_OF_HOUSEHOLD)
                | (filing_status == filing_statuses.SURVIVING_SPOUSE)
            ),
            1,
            where(
                (filing_status == filing_statuses.JOINT)
                & (age_of_spouse > 65),
                2,
                0,  # Default value if none of the conditions are met
            ),
        )

        total_agi = where(
            age_deduction_count > 0, federal_agi + spouse_agi, federal_agi
        )

        total_fda = where(
            (filing_status == filing_statuses.SINGLE)
            | (filing_status == filing_statuses.HEAD_OF_HOUSEHOLD)
            | (filing_status == filing_statuses.SURVIVING_SPOUSE),
            you_fdca,
            you_fdca + spouse_fdca,
        )

        total_fds = where(
            (filing_status == filing_statuses.SINGLE)
            | (filing_status == filing_statuses.HEAD_OF_HOUSEHOLD)
            | (filing_status == filing_statuses.SURVIVING_SPOUSE),
            you_fdcs,
            you_fdcs + spouse_fdcs,
        )

        line4 = total_agi + total_fda
        line6 = line4 - total_fds
        line7 = 0  # Taxable benefits from the tax return, change this.
        line8 = line6 - line7

        threshold_ln9 = where(
            (filing_status == filing_statuses.SINGLE)
            | (filing_status == filing_statuses.HEAD_OF_HOUSEHOLD)
            | (filing_status == filing_statuses.SURVIVING_SPOUSE),
            parameters(
                period
            ).gov.states.va.tax.income.va_age_deduction_threshold.SINGLE,
            where(
                filing_status == filing_statuses.JOINT,
                parameters(
                    period
                ).gov.states.va.tax.income.va_age_deduction_threshold.JOINT,
                parameters(
                    period
                ).gov.states.va.tax.income.va_age_deduction_threshold.SEPARATE,
            ),
        )

        line11 = where(
            line8 >= threshold_ln9,
            line8 - threshold_ln9,
            where(
                (filing_status == filing_statuses.SINGLE)
                | (filing_status == filing_statuses.HEAD_OF_HOUSEHOLD)
                | (filing_status == filing_statuses.SURVIVING_SPOUSE)
                | (filing_status == filing_statuses.SEPARATE),
                12_000,
                where(filing_status == filing_statuses.JOINT, 24_000, 0),
            ),
        )

        line12 = age_deduction_count * 12_000

        line14 = where(line11 < line12, line12 - line11, 0)

        result = where(
            (filing_status == filing_statuses.SINGLE)
            | (filing_status == filing_statuses.HEAD_OF_HOUSEHOLD)
            | (filing_status == filing_statuses.SURVIVING_SPOUSE)
            | (
                (filing_status == filing_statuses.SEPARATE)
                & (age_of_spouse < 65)
            )
            | (
                (filing_status == filing_statuses.JOINT) & (age_of_spouse < 65)
            ),
            line14,
            where(
                (filing_status == filing_statuses.SEPARATE)
                & (age_of_spouse >= 65),
                line14 / 2,
                0,
            ),
        )

        # Check if age_of_head is less than 65 and set result to 0 in that case
        result = where(age_of_head < 65, 0, result)

        return result



from fiscalsim_us.model_api import *


class sc_young_child_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "South Carolina young child exemption"
    unit = USD
    definition_period = YEAR
    reference = "https://dor.sc.gov/forms-site/Forms/SC1040_2022.pdf"
    defined_for = StateCode.SC

    def formula(tax_unit, period, parameters):
        # Get relevant parameter subtree.
        p = parameters(period).gov.states.sc.tax.income.exemptions.young_child
        # Determine eligibility for each person in the tax unit.
        person = tax_unit.members
        meets_age_limit = person("age", period) < p.ineligible_age
        eligible = meets_age_limit & person("is_tax_unit_dependent", period)
        # Count number of eligible people in the tax unit.
        count_eligible = tax_unit.sum(eligible)
        # Multiply by the amount per exemption.
        return count_eligible * p.amount



from fiscalsim_us.model_api import *


class tanf_countable_income(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = YEAR
    label = "TANF countable income"
    documentation = "Countable income for calculating Temporary Assistance for Needy Families benefit."
    unit = USD
    reference = (
        # California:
        "https://ehsd.org/benefits/calworks-cash-aid/calworks-fact-sheet"
    )

    def formula(spm_unit, period, parameters):
        deductions = parameters(
            period
        ).gov.hhs.tanf.cash.amount.countable_income.deductions
        # Annualize the monthly household deduction.
        state = spm_unit.household("state_code_str", period)
        household_deduction = deductions.household[state] * 12
        # First subtract household deduction from unearned income.
        unearned_income = spm_unit("tanf_gross_unearned_income", period)
        countable_unearned_income = max_(
            0, unearned_income - household_deduction
        )
        # Then allocate remaining household deduction to earned income.
        remaining_household_deduction = household_deduction - (
            unearned_income - countable_unearned_income
        )
        gross_earned_income = spm_unit("tanf_gross_earned_income", period)
        earnings_after_household_deduction = max_(
            gross_earned_income - remaining_household_deduction, 0
        )
        # Then subtract percent earnings deduction after household deduction.
        percent_earnings_deduction = deductions.earnings.percent[state]
        total_percent_deduction = (
            earnings_after_household_deduction * percent_earnings_deduction
        )
        countable_earned_income = (
            earnings_after_household_deduction - total_percent_deduction
        )
        return countable_earned_income + countable_unearned_income



from fiscalsim_us.model_api import *


class mn_other_refundable_credits(Variable):
    """
    Other refundable credits that haven't been implemented in other vars.
    Other refundable credits from Line 22 on Minnesota 2022

    * K-12 Education Credit
    * Credit for Parents of Stillborn Children
    * Refundable credit for taxes paid to Wisconsin
    * Credit for Historic Structure Rehabilitation
    * Enterprise Zone Credit
    * Angel Investment Credit
    * Pass Through Entity Credit
    * Claim of right

    If another variable gets added, remove from the list.
    """

    value_type = float
    entity = TaxUnit
    label = "Minnesota other refundable income tax credits"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1c_21.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1c_22.pdf"
    )
    defined_for = StateCode.MN



from fiscalsim_us.model_api import *


class mt_taxable_income(Variable):
    """
    Line 17 on Montana state individual tax return form 2
    """

    value_type = float
    entity = TaxUnit
    label = "Montana taxable income"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT


def formula(tax_unit, period, parameters):
    std_ded = tax_unit("mt_standard_deduction", period)
    itm_ded = tax_unit("mt_itemized_deductions", period)
    deductions = where(itm_ded > std_ded, itm_ded, std_ded)
    exemptions = tax_unit("mt_exemptions", period)
    return max(
        0,
        tax_unit("mt_agi", period) - deductions - exemptions,
    )



from fiscalsim_us.model_api import *


class sep_simple_qualified_plan_contributions(Variable):
    value_type = float
    entity = Person
    label = "SEP, SIMPLE or Qualified plan contributions"
    unit = USD
    definition_period = YEAR



from fiscalsim_us.model_api import *


class household_state_income_tax(Variable):
    # This definition contains all modelled states, and exists to solve circular dependencies in state_income_tax.
    value_type = float
    entity = TaxUnit
    label = "household State tax"
    unit = USD
    definition_period = YEAR
    adds = [
        "ca_income_tax_before_refundable_credits",
        "co_income_tax_before_refundable_credits",
        "dc_income_tax_before_refundable_credits",
        "ia_income_tax_before_refundable_credits",
        "il_total_tax",
        "in_income_tax_before_refundable_credits",
        "ks_income_tax_before_refundable_credits",
        "ky_income_tax_before_refundable_credits",
        "la_income_tax_before_refundable_credits",
        "me_income_tax_before_refundable_credits",
        "ma_income_tax_before_refundable_credits",
        "md_income_tax_before_refundable_credits",
        "mn_income_tax_before_refundable_credits",
        "mt_income_tax_before_refundable_credits",
        "mo_income_tax_before_refundable_credits",
        "nc_income_tax",
        "nd_income_tax_before_refundable_credits",
        "ne_income_tax_before_refundable_credits",
        "nh_income_tax_before_refundable_credits",
        "nj_income_tax_before_refundable_credits",
        "nm_income_tax_before_refundable_credits",
        "ny_income_tax_before_refundable_credits",
        "ok_income_tax_before_refundable_credits",
        "or_income_tax_before_refundable_credits",
        "pa_income_tax",
        "ri_income_tax_before_refundable_credits",
        "sc_income_tax_before_refundable_credits",
        "ut_income_tax_before_refundable_credits",
        "va_income_tax_before_refundable_credits",
        "vt_income_tax_before_refundable_credits",
        "wa_income_tax_before_refundable_credits",
        "wi_income_tax_before_refundable_credits",
        "nyc_income_tax_before_refundable_credits",
    ]
    subtracts = [
        "ca_refundable_credits",  # California
        "co_refundable_credits",  # Colorado
        "dc_refundable_credits",  # District of Columbia
        "ia_refundable_credits",  # Iowa
        "il_refundable_credits",  # Illinois
        "in_refundable_credits",  # Indiana
        "ks_refundable_credits",  # Kansas
        "ky_refundable_credits",  # Kentucky
        "la_refundable_credits",  # Louisiana
        "ma_refundable_credits",  # Massachusetts
        "me_refundable_credits",  # Maine
        "md_refundable_credits",  # Maryland
        "mn_refundable_credits",  # Minnesota
        "mt_refundable_credits",  # Montana
        "mo_refundable_credits",  # Missouri
        # Skip NC, which has no refundable credits
        "nd_refundable_credits",  # North Dakota
        "ne_refundable_credits",  # Nebraska
        "nh_refundable_credits",  # New Hampshire
        "nj_refundable_credits",  # New Jersey
        "nm_refundable_credits",  # New Mexico
        "ny_refundable_credits",  # New York
        "ok_refundable_credits",  # Oklahoma
        "or_refundable_credits",  # Oregon
        # Skip PA, which has no refundable credits
        "ri_refundable_credits",  # Rhode Island
        "sc_refundable_credits",  # South Carolina
        "ut_refundable_credits",  # Utah
        "va_refundable_credits",  # Virginia
        "vt_refundable_credits",  # Vermont
        "wa_refundable_credits",  # Washington
        "wi_refundable_credits",  # Wisconsin
        # LOCAL
        "nyc_refundable_credits",  # New York City
    ]

    def formula(tax_unit, period, parameters):
        if parameters(period).simulation.reported_state_income_tax:
            spm_unit = tax_unit.spm_unit
            total_tax_unit_heads = add(spm_unit, period, ["is_tax_unit_head"])
            spm_unit_state_tax = spm_unit(
                "spm_unit_state_tax_reported", period
            )
            return where(
                total_tax_unit_heads > 0,
                spm_unit_state_tax / total_tax_unit_heads,
                0,
            )
        else:
            return add(
                tax_unit, period, household_state_income_tax.adds
            ) - add(tax_unit, period, household_state_income_tax.subtracts)



from fiscalsim_us.model_api import *


class wv_public_pension_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "West Virginia public pension subtraction"
    defined_for = StateCode.WV
    unit = USD
    definition_period = YEAR
    reference = "https://code.wvlegislature.gov/11-21-12/"

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        # "benefits received under any federal retirement system to which Title 4 U.S.C. §111 applies"
        # https://www.law.cornell.edu/uscode/text/4/111
        federal_pension_income = person(
            "taxable_federal_pension_income", period
        )
        p = parameters(
            period
        ).gov.states.wv.tax.income.subtractions.public_pension
        person_capped = min_(federal_pension_income, p.max_amount)
        # Only applies to head and spouse.
        head_spouse_capped = person_capped * ~person(
            "is_tax_unit_dependent", period
        )
        return tax_unit.sum(head_spouse_capped)



import pandas as pd
import plotly.express as px
from policyengine_core.charts import format_fig, BLUE, GRAY, DARK_GRAY
import streamlit as st

training_log_puf_extended_cps = pd.read_csv(
    "training_log.csv.gz", compression="gzip"
)
training_log_cps = pd.read_csv("training_log_cps.csv.gz", compression="gzip")
calibration_final_results = pd.read_csv(
    "calibration_final_results.csv.gz", compression="gzip"
)

training_log_puf_extended_cps["Source dataset"] = "PUF-extended CPS"
training_log_cps["Source dataset"] = "CPS"

training_log = pd.concat([training_log_puf_extended_cps, training_log_cps])
training_log_targets = training_log.copy()
training_log_targets["value"] = training_log_targets["target"]
training_log_targets["Source dataset"] = "Official"
training_log = pd.concat([training_log, training_log_targets])

name = st.selectbox(
    "Metric",
    training_log.name.unique(),
)

total_loss_log = training_log[training_log.name == name]

fig = px.line(
    total_loss_log,
    x="epoch",
    y="value",
    color="Source dataset",
    color_discrete_map={
        "PUF-extended CPS": BLUE,
        "CPS": GRAY,
        "Official": DARK_GRAY,
    },
)

fig = format_fig(fig)
fig.update_layout(
    title=f"{name} during reweighting, by source dataset",
    yaxis_title="Relative loss change",
    xaxis_title="Epoch",
    legend_title="",
)

# Add annotations for final values
for source_dataset in ["PUF-extended CPS", "CPS", "Official"]:
    final_value = total_loss_log[
        total_loss_log["Source dataset"] == source_dataset
    ].iloc[-1]["value"]
    fig.add_annotation(
        x=total_loss_log["epoch"].max(),
        y=final_value,
        text=f"{final_value:,.0f}",
        arrowhead=1,
        yshift=10,
        ax=0,
        ay=-20,
        bgcolor="white",
    )

fig


def plot_metric(name):
    fig = px.bar(
        calibration_final_results[
            calibration_final_results.Variable == name
        ].sort_values("Value"),
        x="Source dataset",
        y="Value",
        color_discrete_map={
            "Enhanced CPS": BLUE,
            "Calibrated CPS": GRAY,
            "CPS": GRAY,
            "Official": DARK_GRAY,
        },
        color="Source dataset",
    )
    fig = format_fig(fig).update_layout(
        title=f"{name} after reweighting, by source dataset",
    )
    return fig


st.plotly_chart(plot_metric(name))



from fiscalsim_us.model_api import *


class oh_section_179_expense_add_back(Variable):
    value_type = float
    entity = TaxUnit
    label = "OH Section 179 Expense Add Back"
    definition_period = YEAR
    documentation = ""
    reference = ""
    # use federal variables if they are added later



from policyengine_core.data import Dataset
from typing import Type
from pathlib import Path
from fiscalsim_us.data.storage import STORAGE_FOLDER


class UpratedCPS(Dataset):
    data_format = Dataset.ARRAYS

    @staticmethod
    def from_dataset(
        dataset: Type[Dataset],
        out_year: int = 2022,
        new_name: str = None,
        new_label: str = None,
        new_file_path: str = None,
        new_url: str = None,
    ):
        class UpratedCPSFromDataset(UpratedCPS):
            name = new_name or f"{dataset.name}_uprated_{out_year}"
            label = new_label or f"{dataset.label}"
            input_dataset = dataset
            time_period = out_year
            file_path = new_file_path or (
                STORAGE_FOLDER / f"{dataset.name}_uprated_{out_year}.h5"
            )
            url = new_url

        return UpratedCPSFromDataset

    def generate(self):
        from fiscalsim_us import Microsimulation

        input_dataset = self.input_dataset()
        simulation = Microsimulation(dataset=self.input_dataset)

        data = {}
        for variable in input_dataset.variables:
            try:
                data[variable] = simulation.calculate(
                    variable, period=self.output_year
                ).values
            except:
                data[variable] = input_dataset.load(variable)

        self.save_dataset(data)



from fiscalsim_us.model_api import *


class is_optional_senior_or_disabled_for_medicaid(Variable):
    value_type = bool
    entity = Person
    label = "Seniors or disabled people not meeting SSI rules"
    documentation = "Whether this person can claim Medicaid through the State's optional pathway for seniors or people with disabilities."
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/42/1396a#m"

    def formula(person, period, parameters):
        personal_income = person("ssi_countable_income", period)
        personal_assets = person("ssi_countable_resources", period)
        tax_unit = person.tax_unit
        income = tax_unit.sum(personal_income)
        assets = tax_unit.sum(personal_assets)
        ma = parameters(
            period
        ).gov.hhs.medicaid.eligibility.categories.senior_or_disabled
        is_senior_or_disabled = person("is_ssi_aged_blind_disabled", period)
        is_joint = person.tax_unit("tax_unit_is_joint", period)
        state = person.household("state_code_str", period)
        income_limit = where(
            is_joint,
            ma.income.limit.couple[state] * MONTHS_IN_YEAR,
            ma.income.limit.individual[state] * MONTHS_IN_YEAR,
        )
        income_disregard = where(
            is_joint,
            ma.income.disregard.couple[state] * MONTHS_IN_YEAR,
            ma.income.disregard.individual[state] * MONTHS_IN_YEAR,
        )
        asset_limit = where(
            is_joint,
            ma.assets.limit.couple[state],
            ma.assets.limit.individual[state],
        )
        under_limits = (income - income_disregard < income_limit) & (
            assets < asset_limit
        )
        return is_senior_or_disabled & under_limits



from fiscalsim_us.model_api import *


class retired_from_ky_government(Variable):
    value_type = bool
    entity = Person
    label = "Retired from state or local government in Kentucky"
    definition_period = YEAR
    defined_for = StateCode.KY
    default_value = False



from fiscalsim_us.model_api import *


class medical_expense(Variable):
    value_type = float
    entity = Person
    label = "Total medical expenses"
    unit = USD
    definition_period = YEAR

    adds = ["health_insurance_premiums", "medical_out_of_pocket_expenses"]



from fiscalsim_us.model_api import *


class mt_medical_premiums_deduction(Variable):
    """
    Line 2 on itemized deductions schedule
    Medical insurance premiums not deducted elsewhere on the return
    """

    value_type = float
    entity = TaxUnit
    label = "Montana medical insurance premiums"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT



from fiscalsim_us.model_api import *


class nm_low_income_comprehensive_tax_rebate_exemptions(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Mexico low income comprehensive tax rebate exemptions"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://casetext.com/statute/new-mexico-statutes-1978/chapter-7-taxation/article-2-income-tax-general-provisions/section-7-2-14-low-income-comprehensive-tax-rebate?sort=relevance&type=regulation&tab=keyword&jxs=&resultsNav=false",
    )
    defined_for = StateCode.NM

    def formula(tax_unit, period, parameters):
        """
        Quoting from N.M. Stat. § 7-2-14(C):

        For the purposes of this section, the total number of exemptions for which a
        tax rebate may be claimed or allowed is determined by adding the number of
        federal exemptions allowable for federal income tax purposes for each individual
        included in the return who is domiciled in New Mexico...
        """
        federal_exemptions = tax_unit("exemptions", period)
        """
        ...plus two additional exemptions for each individual domiciled in New Mexico
        included in the return who is sixty-five years of age or older...

        NB: The tax form shows that this and the blind exemption only apply to head
        and spouse, not dependents.
        https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf#page=58
        """
        p = parameters(
            period
        ).gov.states.nm.tax.income.rebates.low_income.exemptions
        aged_exemptions_head = p.aged.calc(tax_unit("age_head", period))
        aged_exemptions_spouse = p.aged.calc(tax_unit("age_spouse", period))
        """
        ...plus one additional exemption for each individual domiciled in New Mexico
        included in the return who, for federal income tax purposes, is blind...
        """
        blind_head_plus_spouse = add(
            tax_unit, period, ["blind_head", "blind_spouse"]
        )
        blind_exemptions = p.blind * blind_head_plus_spouse
        """
        ...plus one exemption for each minor child or stepchild of the resident who would
        be a dependent for federal income tax purposes if the public assistance contributing
        to the support of the child or stepchild was considered to have been contributed by
        the resident.

        NB: This section is not defined in the tax form, so we skip it.
        """
        return (
            federal_exemptions
            + aged_exemptions_head
            + aged_exemptions_spouse
            + blind_exemptions
        )



from fiscalsim_us.model_api import *


class mn_home_mortgage_interest_deduction(Variable):
    """
    Other itemized deductions from for M1SA lines 11 and 12
    Can only deduct interest from $750,000-$1,000,000 of the loan
    depending on when the loan was take out
    """

    value_type = float
    entity = TaxUnit
    label = "MN home mortgage interest deduction"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MN
    # use federal variable once it is made



from fiscalsim_us.model_api import *


class ar_inflation_relief_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arkansas inflation relief income-tax credit"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.AR

    def formula(tax_unit, period, parameters):
        # the Arkansas inflationary relief tax credit is calculated based on net taxable income
        # mentioned in individual income tax return form AR1000F line 28
        income = tax_unit("ar_taxable_income", period)
        p = parameters(
            period
        ).gov.states.ar.tax.income.credits.inflationary_relief
        filing_status = tax_unit("filing_status", period)
        max_amount = p.max_amount[filing_status]
        reduction_start = p.reduction.start[filing_status]
        increment = p.reduction.increment[filing_status]
        reduction_per_increment = p.reduction.amount[filing_status]
        excess = max_(income - reduction_start, 0)
        increments = np.ceil(excess / increment)
        total_reduction_amount = increments * reduction_per_increment
        return max_(max_amount - total_reduction_amount, 0)



from fiscalsim_us.model_api import *


class social_security_disability(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    label = "Social Security disability benefits (SSDI)"
    unit = USD



from fiscalsim_us.model_api import *


class mt_charitable_carryover_deduction(Variable):
    """
    Line 13 on Montana itemized deductions schedule
    """

    value_type = float
    entity = TaxUnit
    label = (
        "Montana charitable carryover donation from previous year deduction"
    )
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT



from fiscalsim_us.model_api import *


class ut_apportionable_nonrefundable_credits(Variable):
    """
    Line 24 on Utah 2022 Individual Income Tax return form TC-40. These
    credits include the following categories and are listed at
    https://incometax.utah.gov/credits with their accompanying codes.
    * 04: Capital gain transactions credit
    * 18: Retirement credit
    * 20: Utah my529 credit
    * 23: Health benefit plan credit
    * 26: Gold and silver coin sale credit
    * AH: Social Security benefits credit
    * AJ: Military retirement credit
    * AM: Earned income tax credit
    """

    value_type = float
    entity = TaxUnit
    label = "UT apportionable nonrefundable credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.UT



from policyengine_core.data import Dataset
from typing import Type
import pandas as pd
from pathlib import Path
from fiscalsim_us.data.storage import STORAGE_FOLDER
import os


class PovertyTracker(Dataset):
    name = "poverty_tracker"
    label = "Poverty Tracker"
    data_format = Dataset.ARRAYS
    time_period = "2023"
    file_path = STORAGE_FOLDER / "poverty_tracker.h5"

    # Fetch the secret URL from the environment variable.
    # Replace with a valid URL if running locally.
    POVERTY_TRACKER_URL = os.environ.get("POVERTYTRACKER_RAW_URL")

    def generate(self):
        household_df = self.load_pt_data()
        situations = household_df.apply(get_hh_situtation, axis=1)

        if not self.POVERTY_TRACKER_URL:
            raise ValueError(
                "The POVERTY_TRACKER_URL environment variable is not set!"
            )

        person_ids = []
        person_group_ids = []
        group_ids = []

        data = {
            "age": [],
            "employment_income": [],
        }

        current_person_id = 0
        current_group_id = 0
        group_ids = household_df["public_id"].values
        for situation in situations:
            for person in situation["people"]:
                person_ids.append(current_person_id)
                person_group_ids.append(group_ids[current_group_id])

                age = situation["people"][person]["age"]
                employment_income = situation["people"][person][
                    "employment_income"
                ]

                # Clarify why age is sometimes directly in situation["people"][person]["age"] and sometimes in situation["people"][person]["age"][TAX_YEAR_STR]
                data["age"].append(
                    age.get(TAX_YEAR_STR) if isinstance(age, dict) else age
                )
                data["employment_income"].append(
                    employment_income.get(TAX_YEAR_STR)
                    if isinstance(employment_income, dict)
                    else employment_income
                )

                current_person_id += 1

            current_group_id += 1

        data["in_nyc"] = [True] * len(group_ids)
        data["state_name"] = ["NY"] * len(group_ids)

        group_id_names = [
            "tax_unit",
            "marital_unit",
            "spm_unit",
            "household",
            "family",
        ]

        for group_id_name in group_id_names:
            data[group_id_name + "_id"] = group_ids
            data["person_" + group_id_name + "_id"] = person_group_ids

        data["person_id"] = person_ids

        self.save_dataset(data)

    def load_pt_data(
        self,
        url: str = POVERTY_TRACKER_URL,
    ) -> pd.DataFrame:
        """
        Loads household-level raw PovertyTracker dataset.
        """
        # Download raw data from PovertyTracker website.
        pt = pd.read_csv(url)

        columns_to_rename = {
            # "imp_qearnhd_tc": "hh_head_earnings", # not currently used
            # "imp_qearnsp_tc": "spouse_earnings", # not currently used
            "imp_qage_tc": "age",
            "qopmres": "income",
        }

        return pt.rename(columns=columns_to_rename)


def extract_hh_data(observation: pd.Series) -> dict:
    """
    Takes in a row of the PovertyTracker dataframe and extracts
    information on relatives living with the household head.
    """
    relatives_dict = {}

    # If someone(s) else lives with respondent, q16a4 = 1.
    if observation["q16a4"] != 1:
        return relatives_dict

    var_number = 1
    relative_var = "q16a4_rel" + str(var_number)
    dependent_counter = 1
    while pd.notnull(relative_var) and var_number < 10:
        if observation[relative_var] == 1:  # Spouse
            partner_age = observation[f"q16a4_age{var_number}_tc"]
            relatives_dict["your partner"] = {
                "age": partner_age if partner_age != 970 else 40
            }

        elif observation[relative_var] in [
            3,
            4,
            5,
            7,
        ]:  # child, stepchild, grandchild, or foster child
            relative_age = observation[f"q16a4_age{var_number}_tc"]
            if relative_age < 18:  # assume only children are dependents
                relatives_dict[f"dependent{dependent_counter}"] = {
                    "age": relative_age if relative_age != 970 else 40
                }
            dependent_counter += 1

        var_number += 1

    return relatives_dict


TAX_YEAR = 2023
TAX_YEAR_STR = str(TAX_YEAR)


def get_hh_situtation(row: pd.Series):
    situation_dict = {
        "people": {
            "you": {
                "age": {TAX_YEAR_STR: row["age"]},
                "employment_income": {TAX_YEAR_STR: row["income"]},
            }
        },
        "families": {"your family": {"members": ["you"]}},
        "marital_units": {},
        "tax_units": {"your tax unit": {"members": ["you"]}},
        "spm_units": {
            "your household": {
                "members": [
                    "you",
                    # "your partner",
                    # "your first dependent"
                ]
            }
        },
        "households": {
            "your household": {
                "members": ["you"],
                "state_name": {TAX_YEAR_STR: "NY"},
                "in_nyc": True,
            }
        },
    }

    hh_data = extract_hh_data(row)
    situation_dict["people"].update(hh_data)

    # Add spouse and dependents to families, marital_units, tax_units, spm_units
    family_members = ["you"]
    marital_unit_members = ["you"]
    for member, member_data in hh_data.items():
        family_members.append(member)
        marital_unit_members.append(member)
        situation_dict["people"][member].update(
            {"employment_income": {TAX_YEAR_STR: 0}}
        )
        situation_dict["tax_units"]["your tax unit"]["members"].append(member)
        situation_dict["spm_units"]["your household"]["members"].append(member)
        situation_dict["households"]["your household"]["members"].append(
            member
        )

        # marital_unit_id = "{}'s marital unit".format(member)
        # situation_dict["marital_units"][marital_unit_id] = {
        #     "members": [member]
        # }

        # if member.startswith("dependent"):
        #     situation_dict["marital_units"][marital_unit_id]["marital_unit_id"] = {
        #         TAX_YEAR_STR: dependent_counter
        #     }
        #     dependent_counter += 1

    situation_dict["families"]["your family"]["members"] = family_members
    situation_dict["marital_units"]["your marital unit"] = {
        "members": marital_unit_members
    }

    return situation_dict



import pytest


@pytest.mark.skipif(True, reason="This test temporarily suspended.")
def test_microsim_runs_cps():
    import numpy as np
    from fiscalsim_us import Microsimulation

    sim = Microsimulation()
    hnet = sim.calc("household_net_income")
    assert not hnet.isna().any(), "Some households have NaN net income."
    hidecile = sim.calc("household_income_decile")
    assert np.all(hidecile >= 1) and np.all(hidecile <= 10)
    sidecile = sim.calc("spm_unit_income_decile")
    assert np.all(sidecile >= 1) and np.all(sidecile <= 10)
    idecile = sim.calc("income_decile")
    assert np.all(idecile >= 1) and np.all(idecile <= 10)



from fiscalsim_us.model_api import *


class ecpa_adult_dependent_credit(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    unit = USD
    label = "End Child Poverty Act Adult Dependent Credit"

    def formula_2022(tax_unit, period, parameters):
        person = tax_unit.members
        p = parameters(
            period
        ).gov.contrib.congress.tlaib.end_child_poverty_act.adult_dependent_credit
        # Adult dependent credit.
        dependent = person("is_tax_unit_dependent", period)
        adult = person("age", period) >= p.min_age
        return p.amount * tax_unit.sum(adult & dependent)



from fiscalsim_us.model_api import *


class wa_income_tax_before_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Washington income tax before refundable credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.WA
    adds = ["wa_capital_gains_tax"]



from fiscalsim_us.model_api import *


class c04600(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = "Personal exemptions after phase-out"
    unit = USD

    def formula(tax_unit, period, parameters):
        phase_out = parameters(period).gov.irs.income.exemption.phase_out
        phase_out_start = tax_unit("exemption_phase_out_start", period)
        line_5 = max_(
            0, tax_unit("adjusted_gross_income", period) - phase_out_start
        )
        line_6 = line_5 / (2500 / tax_unit("sep", period))
        line_7 = phase_out.rate * line_6
        return tax_unit("pre_c04600", period) * (1 - line_7)



from fiscalsim_us.model_api import *


class is_medicaid_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Eligible for Medicaid"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/42/1396a#a_10"

    def formula(person, period, parameters):
        category = person("medicaid_category", period)
        return category != category.possible_values.NONE



from fiscalsim_us.model_api import *


class tax_unit_taxable_social_security(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Taxable Social Security benefits"
    documentation = "Social security (OASDI) benefits included in AGI, including tier 1 railroad retirement benefits."
    unit = USD
    reference = "https://www.law.cornell.edu/uscode/text/26/86"

    def formula(tax_unit, period, parameters):
        ss = parameters(period).gov.irs.social_security.taxability
        gross_ss = tax_unit("tax_unit_social_security", period)

        # The legislation directs the usage an income definition that is
        # a particularly modified AGI, plus half of gross Social Security
        # payments. We assume that the 'half' here is the same underlying
        # parameter as the lower taxability marginal rate (also 50% in the
        # baseline), and that they would be mechanically the same parameter.

        ss_fraction = ss.rate.lower * gross_ss
        modified_agi_plus_half_ss = (
            tax_unit("taxable_ss_magi", period) + ss_fraction
        )
        filing_status = tax_unit("filing_status", period)

        base_amount = ss.threshold.lower[filing_status]
        adjusted_base_amount = ss.threshold.upper[filing_status]

        under_first_threshold = modified_agi_plus_half_ss < base_amount
        under_second_threshold = (
            modified_agi_plus_half_ss < adjusted_base_amount
        )

        excess_over_base = max_(0, modified_agi_plus_half_ss - base_amount)
        excess_over_adjusted_base = max_(
            0, modified_agi_plus_half_ss - adjusted_base_amount
        )

        amount_if_under_second_threshold = ss.rate.lower * min_(
            excess_over_base, gross_ss
        )
        amount_if_over_second_threshold = min_(
            ss.rate.upper * excess_over_adjusted_base
            + min_(
                amount_if_under_second_threshold,
                ss.rate.lower * (adjusted_base_amount - base_amount),
            ),
            ss.rate.upper * gross_ss,
        )
        return select(
            [
                under_first_threshold,
                under_second_threshold,
                True,
            ],
            [
                0,
                amount_if_under_second_threshold,
                amount_if_over_second_threshold,
            ],
        )



from fiscalsim_us.model_api import *


class ut_taxpayer_credit_reduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "UT taxpayer credit reduction"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.UT

    def formula(tax_unit, period, parameters):
        gov = parameters(period).gov
        ut_taxpayer_credit_params_path = (
            gov.states.ut.tax.income.credits.taxpayer_credit
        )
        filing_status = tax_unit("filing_status", period)
        phase_out_start = ut_taxpayer_credit_params_path.phase_out.start[
            filing_status
        ]
        phase_out_rate = ut_taxpayer_credit_params_path.phase_out.rate
        ut_taxable_income = tax_unit("ut_taxable_income", period)
        income_subject_to_phase_out = max_(
            ut_taxable_income - phase_out_start, 0
        )
        return income_subject_to_phase_out * phase_out_rate



from fiscalsim_us.model_api import *


class mt_medical_dental_expenses_deduction(Variable):
    """
    Line 1 on itemized deductions schedule
    """

    value_type = float
    entity = TaxUnit
    label = "Montana medical and dental expense deduction"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.mt.tax.income.deductions.itemized.medical_dental
        line1a = tax_unit("mt_medical_dental_expense", period)
        line1b = tax_unit("mt_agi", period)
        line1c = line1b * p.rate

        return max_(line1c - line1a, 0)



from fiscalsim_us.model_api import *


class nyc_school_tax_credit_fixed_amount(Variable):
    value_type = float
    entity = TaxUnit
    label = "NYC School Tax Credit Fixed Amount"
    unit = USD
    definition_period = YEAR
    defined_for = "in_nyc"

    def formula(tax_unit, period, parameters):
        # First get their NYAGI.
        income = tax_unit("nyc_school_credit_income", period)

        # Then get their filing status.
        filing_status = tax_unit("filing_status", period)

        # Then get the School Tax Credit fixed amount part of the parameter tree.
        p = parameters(period).gov.local.ny.nyc.tax.income.credits.school.fixed

        # Calculate eligibility.
        eligible = income <= p.income_limit

        # Calculate amount if eligible, which varies only with filing status.
        amount_if_eligible = p.amount[filing_status]

        # Return calculated amount.
        return eligible * amount_if_eligible



from fiscalsim_us.model_api import *


class household_benefits(Variable):
    value_type = float
    entity = Household
    label = "benefits"
    unit = USD
    definition_period = YEAR
    adds = [
        "social_security",
        "ssi",
        "state_supplement",
        "co_state_supplement",
        "co_oap",
        "ca_cvrp",  # California Clean Vehicle Rebate Project.
        "ca_care",
        "ca_fera",
        "snap",
        "wic",
        "free_school_meals",
        "reduced_price_school_meals",
        "lifeline",
        "acp",
        "ebb",
        "tanf",
        "high_efficiency_electric_home_rebate",
        "residential_efficiency_electrification_rebate",
        "unemployment_compensation",
        # Contributed.
        "basic_income",
        "spm_unit_capped_housing_subsidy",
    ]



from fiscalsim_us.model_api import *


class hi_subtractions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Hawaii subtractions from federal adjusted gross income"
    reference = "https://files.hawaii.gov/tax/forms/2022/n11ins.pdf#page=13"
    defined_for = StateCode.HI
    unit = USD
    definition_period = YEAR
    adds = "gov.states.hi.tax.income.subtractions"



from fiscalsim_us.model_api import *


class va_subtractions_to_federal_agi(Variable):
    value_type = float
    entity = TaxUnit
    label = "VA subtractions to federal agi https://www.tax.virginia.gov/sites/default/files/taxforms/individual-income-tax/2021/schedule-2021.pdf"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.VA

    def formula(tax_unit, period, parameters):
        income_from_obligations_state_exempt = tax_unit(
            "va_income_from_obligations_state_exempt", period
        )

        fixed_date_conformity_subtractions = tax_unit(
            "va_fixed_date_conformity_subtractions", period
        )

        disability_income_reported_as_wages = tax_unit(
            "va_disability_income_reported_as_wages", period
        )

        subtractions_to_fed_agi = (
            income_from_obligations_state_exempt
            + fixed_date_conformity_subtractions
            + disability_income_reported_as_wages
        )

        return subtractions_to_fed_agi



from fiscalsim_us.model_api import *


class physician_services_expense(Variable):
    value_type = float
    entity = Person
    label = "Physician services expenses"
    unit = USD
    definition_period = YEAR



from fiscalsim_us.model_api import *


class eitc_child_count(Variable):
    value_type = int
    entity = TaxUnit
    label = "EITC-qualifying children"
    unit = USD
    documentation = "Number of children qualifying as children for the EITC."
    definition_period = YEAR

    adds = ["is_eitc_qualifying_child"]



from fiscalsim_us.model_api import *


class mn_personal_property_tax(Variable):
    """
    State and local personal propery tax (non real estate)
    that for taxes based on value alone
    Line 10 of 2022 M1SA, Minnesota Itemized Deductions
    Line 5c of Federal Schedule A
    """

    value_type = float
    entity = TaxUnit
    label = "MN personal property tax"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MN



from fiscalsim_us.model_api import *


class qualified_business_income(Variable):
    value_type = float
    entity = Person
    label = "Qualified business income"
    documentation = "Business income that qualifies for the qualified business income deduction."
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/199A#c"

    def formula(person, period, parameters):
        p = parameters(period).gov.irs.deductions
        income_components = p.qbi.income_definition
        gross_qbi = add(person, period, income_components)
        deduction_components = p.qbi.deduction_definition
        qbi_deductions = add(person, period, deduction_components)
        adjusted_qbi = max_(0, gross_qbi - qbi_deductions)
        qualified = person("business_is_qualified", period)
        return adjusted_qbi * qualified



from fiscalsim_us.model_api import *


class mn_other_itemized_deductions(Variable):
    """
    Other itemized deductions from for M1SA lines 20 and 24
    * Federal estate tax on income in respect of a decedent
    * Deduction for amortizable bond premium
    * Deduction for repayment of amounts under a claim of right
    * Certain unrecovered investment in pension
    * Impairment-related work expenses of a person with a disability
    * Deduction allowable in connection with personal property used
    in a short sale as described under section 67(b)(8)
    """

    value_type = float
    entity = TaxUnit
    label = "MN other itemized deductions"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MN



from .cps import (
    CPS_2020,
    CPS_2021,
    CPS_2022,
    CPS_2023,
)

from .raw_cps import RawCPS_2020, RawCPS_2021, RawCPS_2022

from .enhanced_cps import (
    EnhancedCPS_2023,
    PUFExtendedCPS_2023,
    CalibratedCPS_2023,
)



from fiscalsim_us.model_api import *


class in_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Indiana deductions"
    unit = USD
    definition_period = YEAR
    reference = "http://iga.in.gov/legislative/laws/2021/ic/titles/006#6-3"

    adds = [
        "in_renters_deduction",
        "in_homeowners_property_tax_deduction",
        "salt_refund_last_year",
        "us_govt_interest",
        "tax_unit_taxable_social_security",  # includes railroad retirement benefits
        "in_military_service_deduction",
        "in_nonpublic_school_deduction",
        "in_nol",
        "in_unemployment_compensation_deduction",
        "in_other_deductions",
    ]



from fiscalsim_us.model_api import *


def _apply_ssi_exclusions(
    earned_income: ArrayLike,
    unearned_income: ArrayLike,
    parameters: ParameterNode,
    period: Period,
) -> ArrayLike:
    """Applies the SSI exclusions to earned income and unearned income, combining the result.

    Args:
        earned_income (ArrayLike): Earned income for each person.
        unearned_income (ArrayLike): Unearned income for each person.
        parameters (ParameterNode): The root of the parameter tree.

    Returns:
        ArrayLike: SSI countable income.
    """

    earned = earned_income / MONTHS_IN_YEAR
    unearned = unearned_income / MONTHS_IN_YEAR
    exclusions = parameters(period).gov.ssa.ssi.income.exclusions
    # Subtract general exclusion from unearned income first.
    unearned_exclusion = min_(exclusions.general, unearned)
    countable_unearned = unearned - unearned_exclusion
    # Remaining general exclusion is treated as an earned income exclusion.
    remaining_general_exclusion = exclusions.general - unearned_exclusion
    earned_exclusion = exclusions.earned + remaining_general_exclusion
    # Subtract the percentage of earned income above the flat exclusion.
    earned_after_flat_exclusion = max_(earned - earned_exclusion, 0)
    countable_earned_share = 1 - exclusions.earned_share
    countable_earned = earned_after_flat_exclusion * countable_earned_share
    return (countable_unearned + countable_earned) * MONTHS_IN_YEAR



from fiscalsim_us.model_api import *


class ks_exemptions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Kansas exemptions amount"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.ksrevenue.gov/pdf/ip21.pdf"
        "https://www.ksrevenue.gov/pdf/ip22.pdf"
    )
    defined_for = StateCode.KS

    def formula(tax_unit, period, parameters):
        num_exemptions = tax_unit("ks_count_exemptions", period)
        p = parameters(period).gov.states.ks.tax.income.exemptions
        return num_exemptions * p.amount



import logging
from policyengine_core.data import PublicDataset
import h5py
from fiscalsim_us.data.datasets.acs.raw_acs import RawACS
from fiscalsim_us.data.storage import fiscalsim_us_MICRODATA_FOLDER
from pandas import DataFrame


class ACS(PublicDataset):
    name = "acs"
    is_openfisca_compatible = True
    label = "ACS"
    folder_path = fiscalsim_us_MICRODATA_FOLDER

    def generate(self, year: int) -> None:
        """Generates the ACS dataset.

        Args:
            year (int): The year of the raw ACS to use.
        """

        # Prepare raw ACS tables
        year = int(year)
        if year in self.years:
            self.remove(year)
        if year not in RawACS.years:
            RawACS.generate(year)

        raw_data = RawACS.load(year)
        acs = h5py.File(ACS.file(year), mode="w")

        person, spm_unit, household = [
            raw_data[entity] for entity in ("person", "spm_unit", "household")
        ]
        # Add primary and foreign keys

        household.SERIALNO = household.SERIALNO.astype(int)
        person.SERIALNO = person.SERIALNO.astype(int)
        person.SPORDER = person.SPORDER.astype(int)
        person.SPM_ID = person.SPM_ID.astype(int)
        spm_unit.SPM_ID = spm_unit.SPM_ID.astype(int)

        logging.info(
            f"Persons with a linked household {person.SERIALNO.isin(household.SERIALNO).mean():.1%}"
        )
        person = person[person.SERIALNO.isin(household.SERIALNO)]
        logging.info(
            f"Households with a linked person {household.SERIALNO.isin(person.SERIALNO).mean():.1%}"
        )
        household = household[household.SERIALNO.isin(person.SERIALNO)]
        logging.info(
            f"SPM units with a linked person {spm_unit.SPM_ID.isin(person.SPM_ID).mean():.1%}"
        )
        spm_unit = spm_unit[spm_unit.SPM_ID.isin(person.SPM_ID)]

        add_id_variables(acs, person, spm_unit, household)
        add_person_variables(acs, person)
        add_spm_variables(acs, spm_unit)
        add_household_variables(acs, household)

        raw_data.close()
        acs.close()


ACS = ACS()


def add_id_variables(
    acs: h5py.File,
    person: DataFrame,
    spm_unit: DataFrame,
    household: DataFrame,
) -> None:
    """Add basic ID and weight variables.

    Args:
        acs (h5py.File): The ACS dataset file.
        person (DataFrame): The person table of the ACS.
        spm_unit (DataFrame): The SPM unit table created from the person table
            of the ACS.
        household (DataFrame): The household table of the ACS.
    """
    acs["person_id"] = person.SERIALNO * 1e2 + person.SPORDER
    acs["person_spm_unit_id"] = person.SPM_ID
    acs["spm_unit_id"] = spm_unit.SPM_ID
    # ACS doesn't have tax units.
    acs["tax_unit_id"] = spm_unit.SPM_ID
    # Until we add a family table, we'll use the person table.
    acs["family_id"] = spm_unit.SPM_ID
    acs["person_household_id"] = person.SERIALNO
    acs["person_tax_unit_id"] = person.SPM_ID
    acs["person_family_id"] = person.SPM_ID
    acs["household_id"] = household.SERIALNO

    # TODO: add marital unit IDs - using person IDs for now
    acs["person_marital_unit_id"] = person.SERIALNO
    acs["marital_unit_id"] = person.SERIALNO.unique()

    # Add weights
    acs["person_weight"] = person.PWGTP
    acs["household_weight"] = household.WGTP


def add_person_variables(acs: h5py.File, person: DataFrame) -> None:
    acs["age"] = person.AGEP
    acs["employment_income"] = person.WAGP
    acs["self_employment_income"] = person.SEMP
    acs["total_income"] = person.PINCP


def add_spm_variables(acs: h5py.File, spm_unit: DataFrame) -> None:
    acs["spm_unit_net_income_reported"] = spm_unit.SPM_RESOURCES
    acs["spm_unit_spm_threshold"] = spm_unit.SPM_POVTHRESHOLD


def add_household_variables(acs: h5py.File, household: DataFrame) -> None:
    acs["household_vehicles_owned"] = household.VEH
    acs["state_fips"] = acs["household_state_fips"] = household.ST



from fiscalsim_us.model_api import *


class vt_military_retirement_pay_exclusion(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Vermont military retirement income exclusion"
    reference = (
        "https://tax.vermont.gov/individuals/seniors-and-retirees",
        "https://legislature.vermont.gov/statutes/section/32/151/05830e",  # Legal Code Titl. 32 V.S.A. § 5830e (d)
    )
    unit = USD
    defined_for = StateCode.VT
    documentation = (
        "Vermont military retirement benefits exempt from Vermont taxation."
    )

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        p = parameters(
            period
        ).gov.states.vt.tax.income.agi.retirement_income_exemption
        # Get retirement amount from military retirement system
        tax_unit_military_retirement_pay = add(
            tax_unit, period, ["military_retirement_pay"]
        )
        # Retirement income from systems other than social security have maximum amount.
        return min_(
            tax_unit_military_retirement_pay, p.military_retirement.amount
        )



from fiscalsim_us.model_api import *


class mn_adjustmnets_fed6251(Variable):
    """
    Adjustements and preferences to income for alternative
    minimum tax on federal Form 6251
    Line 2 Minnesota 2022 form M1MT, Alternative Minimum Tax
    """

    value_type = float
    entity = TaxUnit
    label = "MN income adjustements fed Form 6251"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MN



from fiscalsim_us.model_api import *


class medicaid_income_level(Variable):
    value_type = float
    entity = Person
    label = "Medicaid income level"
    unit = "/1"
    documentation = "Income for Medicaid as a percentage of the federal poverty line for this person."
    definition_period = YEAR

    def formula(person, period, parameters):
        return person.tax_unit("tax_unit_medicaid_income_level", period)



from fiscalsim_us.model_api import *


class ny_tanf_countable_gross_unearned_income(Variable):
    value_type = float
    entity = SPMUnit
    label = "New York TANF countable gross unearned income"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NY

    adds = "gov.states.ny.otda.tanf.income.unearned"



from io import BytesIO
import logging
from typing import List
from zipfile import ZipFile
import pandas as pd
from policyengine_core.data import PublicDataset
import requests
from tqdm import tqdm
from fiscalsim_us.data.storage import fiscalsim_us_MICRODATA_FOLDER


logging.getLogger().setLevel(logging.INFO)

PERSON_COLUMNS = [
    "SERIALNO",  # Household ID
    "SPORDER",  # Person number within household
    "PWGTP",  # Person weight
    "AGEP",  # Age
    "CIT",  # Citizenship
    "MAR",  # Marital status
    "WAGP",  # Wage/salary
    "SSP",  # Social security income
    "SSIP",  # Supplemental security income
    "SEX",  # Sex
    "SEMP",  # Self-employment income
    "SCHL",  # Educational attainment
    "RETP",  # Retirement income
    "PAP",  # Public assistance income
    "OIP",  # Other income
    "PERNP",  # Total earnings
    "PINCP",  # Total income
    "POVPIP",  # Income-to-poverty line percentage
    "RAC1P",  # Race
]

HOUSEHOLD_COLUMNS = [
    "SERIALNO",  # Household ID
    "PUMA",  # PUMA area code
    "ST",  # State code
    "ADJHSG",  # Adjustment factor for housing dollar amounts
    "ADJINC",  # Adjustment factor for income
    "WGTP",  # Household weight
    "NP",  # Number of persons in household
    "BDSP",  # Number of bedrooms
    "ELEP",  # Electricity monthly cost
    "FULP",  # Fuel monthly cost
    "GASP",  # Gas monthly cost
    "RMSP",  # Number of rooms
    "RNTP",  # Monthly rent
    "TEN",  # Tenure
    "VEH",  # Number of vehicles
    "FINCP",  # Total income
    "GRNTP",  # Gross rent
]


class RawACS(PublicDataset):
    name = "raw_acs"
    label = "Raw ACS"
    is_openfisca_compatible = False
    folder_path = fiscalsim_us_MICRODATA_FOLDER

    def generate(self, year: int) -> None:
        year = int(year)
        if year in self.years:
            self.remove(year)

        spm_url = f"https://www2.census.gov/programs-surveys/supplemental-poverty-measure/datasets/spm/spm_{year}_pu.dta"
        person_url = f"https://www2.census.gov/programs-surveys/acs/data/pums/{year}/1-Year/csv_pus.zip"
        household_url = f"https://www2.census.gov/programs-surveys/acs/data/pums/{year}/1-Year/csv_hus.zip"

        # The data dictionary for 2019 can be found here: https://www2.census.gov/programs-surveys/acs/tech_docs/pums/data_dict/PUMS_Data_Dictionary_2019.pdf

        try:
            with pd.HDFStore(RawACS.file(year)) as storage:
                # Household file
                logging.info(f"Downloading household file")
                household = concat_zipped_csvs(
                    household_url, "psam_hus", HOUSEHOLD_COLUMNS
                )
                # Remove group quarters (zero weight)
                household = household[
                    ~household.SERIALNO.str.contains("2019GQ")
                ]
                household["SERIALNO"] = household["SERIALNO"].apply(
                    lambda x: int(x.replace("2019HU", ""))
                )
                storage["household"] = household
                # Person file
                logging.info(f"Downloading person file")
                person = concat_zipped_csvs(
                    person_url, "psam_pus", PERSON_COLUMNS
                )
                person = person[~person.SERIALNO.str.contains("2019GQ")]
                person["SERIALNO"] = person["SERIALNO"].apply(
                    lambda x: int(x.replace("2019HU", ""))
                )
                storage["person"] = person
                # SPM unit file
                logging.info(f"Downloading SPM unit file")
                spm_person = pd.read_stata(spm_url).fillna(0)
                spm_person.columns = spm_person.columns.str.upper()
                create_spm_unit_table(storage, spm_person)
        except Exception as e:
            RawACS.remove(year)
            logging.error(
                f"Attempted to extract and save the CSV files, but encountered an error: {e}"
            )
            raise e


RawACS = RawACS()


def concat_zipped_csvs(
    url: str, prefix: str, columns: List[str]
) -> pd.DataFrame:
    """Downloads the ACS microdata, which is a zip file containing two halves in CSV format.

    Args:
        url (str): The URL of the data server.
        prefix (str): The prefix of the filenames, before a/b.csv.
        columns (List[str]): The columns to filter (avoids hitting memory limits).

    Returns:
        pd.DataFrame: The concatenated DataFrame.
    """
    req = requests.get(url, stream=True)
    with BytesIO() as f:
        pbar = tqdm()
        for chunk in req.iter_content(chunk_size=1024):
            if chunk:  # filter out keep-alive new chunks
                pbar.update(len(chunk))
                f.write(chunk)
        f.seek(0)
        zf = ZipFile(f)
        logging.info(f"Loading the first half of the dataset")
        a = pd.read_csv(zf.open(prefix + "a.csv"), usecols=columns)
        logging.info(f"Loading the second half of the dataset")
        b = pd.read_csv(zf.open(prefix + "b.csv"), usecols=columns)
    logging.info(f"Concatenating datasets")
    res = pd.concat([a, b]).fillna(0)
    res.columns = res.columns.str.upper()
    return res


def create_spm_unit_table(storage: pd.HDFStore, person: pd.DataFrame) -> None:
    SPM_UNIT_COLUMNS = [
        "CAPHOUSESUB",
        "CAPWKCCXPNS",
        "CHILDCAREXPNS",
        "EITC",
        "ENGVAL",
        "EQUIVSCALE",
        "FEDTAX",
        "FEDTAXBC",
        "FICA",
        "GEOADJ",
        "MEDXPNS",
        "NUMADULTS",
        "NUMKIDS",
        "NUMPER",
        "POOR",
        "POVTHRESHOLD",
        "RESOURCES",
        "SCHLUNCH",
        "SNAPSUB",
        "STTAX",
        "TENMORTSTATUS",
        "TOTVAL",
        "WCOHABIT",
        "WICVAL",
        "WKXPNS",
        "WUI_LT15",
        "ID",
    ]
    spm_table = (
        person[["SPM_" + column for column in SPM_UNIT_COLUMNS]]
        .groupby(person.SPM_ID)
        .first()
    )

    original_person_table = storage["person"]
    # Ensure that join keys are the same type.
    JOIN_COLUMNS = ["SERIALNO", "SPORDER"]
    original_person_table[JOIN_COLUMNS] = original_person_table[
        JOIN_COLUMNS
    ].astype(int)
    person[JOIN_COLUMNS] = person[JOIN_COLUMNS].astype(int)
    # Add SPM_ID from the SPM person table to the original person table.
    combined_person_table = pd.merge(
        original_person_table,
        person[JOIN_COLUMNS + ["SPM_ID"]],
        on=JOIN_COLUMNS,
    )

    storage["person"] = combined_person_table
    storage["spm_unit"] = spm_table



from fiscalsim_us.model_api import *


class self_employment_tax(Variable):
    value_type = float
    entity = Person
    label = "self-employment tax"
    definition_period = YEAR
    unit = USD

    def formula(person, period, parameters):
        if parameters(
            period
        ).gov.contrib.ubi_center.flat_tax.abolish_self_emp_tax:
            return 0
        return person("self_employment_social_security_tax", period) + person(
            "self_employment_medicare_tax", period
        )



from fiscalsim_us.model_api import *


class va_calc_line_14(Variable):
    value_type = float
    entity = TaxUnit
    label = "VA interim step to get to total virginia taxable income, line 14 on form 760 https://www.tax.virginia.gov/sites/default/files/taxforms/individual-income-tax/2021/760-2021.pdf"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.VA

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        filing_statuses = filing_status.possible_values

        va_standard_deduction = where(
            (filing_status == filing_statuses.SINGLE)
            | (filing_status == filing_statuses.SURVIVING_SPOUSE)
            | (filing_status == filing_statuses.HEAD_OF_HOUSEHOLD),
            parameters(
                period
            ).gov.states.va.tax.income.va_standard_deduction.SINGLE,
            where(
                filing_status == filing_statuses.JOINT,
                parameters(
                    period
                ).gov.states.va.tax.income.va_standard_deduction.JOINT,
                where(
                    filing_status == filing_statuses.SEPARATE,
                    parameters(
                        period
                    ).gov.states.va.tax.income.va_standard_deduction.SEPARATE,
                    0,  # Default value if none of the conditions are met
                ),
            ),
        )

        line_12 = tax_unit("va_exemptions", period)

        subtotal = va_standard_deduction + line_12

        return subtotal



from fiscalsim_us.model_api import *


class ut_income_tax_after_credits(Variable):
    """
    Line 22 of Utah 2022 Individual Income Tax return form TC-40.
    """

    value_type = float
    entity = TaxUnit
    label = "UT income tax after standard nonrefundable credits"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        income_tax_before_credits = tax_unit(
            "ut_income_tax_before_credits", period
        )
        standard_non_refundable_credits = tax_unit(
            "ut_taxpayer_credit", period
        )
        return max_(
            income_tax_before_credits - standard_non_refundable_credits, 0
        )



from fiscalsim_us.model_api import *


class foreign_tax_credit(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    documentation = "Foreign tax credit from Form 1116"
    unit = USD



from fiscalsim_us.model_api import *


class md_ctc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maryland Child Tax Credit"
    definition_period = YEAR
    unit = USD
    reference = (
        "https://casetext.com/statute/code-of-maryland/article-tax-general/title-10-income-tax/subtitle-7-income-tax-credits/section-10-751-effective-until-712026-tax-credit-for-qualified-child"
        "https://law.justia.com/codes/maryland/2022/tax-general/title-10/subtitle-7/section-10-751/"
    )
    defined_for = "md_ctc_eligible"

    def formula_2020(tax_unit, period, parameters):
        p = parameters(period).gov.states.md.tax.income.credits.ctc
        person = tax_unit.members
        dependent = person("is_tax_unit_dependent", period)
        disabled = person("is_disabled", period)
        age_limit = where(
            disabled, p.age_threshold.disabled, p.age_threshold.main
        )
        meets_age_limit = person("age", period) < age_limit
        eligible = dependent & meets_age_limit
        eligible_children = tax_unit.sum(eligible)
        md_ctc = eligible_children * p.amount
        if p.reduced_by_federal_credit:
            federal_ctc = tax_unit("ctc", period)
            return max_(md_ctc - federal_ctc, 0)
        return md_ctc



from fiscalsim_us.model_api import *


class de_itemized_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Delaware itemized deductions"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf#page=7"
        "https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf#page=11"
        "https://casetext.com/statute/delaware-code/title-30-state-taxes/part-ii-income-inheritance-and-estate-taxes/chapter-11-personal-income-tax/subchapter-ii-resident-individuals/section-1109-itemized-deductions-for-application-of-this-section-see-66-del-laws-c-86-section-8"
    )
    defined_for = StateCode.DE

    adds = ["itemized_deductions_less_salt"]
    # Per Law: Self employed filers can deduct the health insurance premiums
    # less the amount allowed as a deduction on the federal return
    # We omit this because it is not on the tax form.



from fiscalsim_us.model_api import *


class oh_federal_conformity_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "OH federal conformity deductions"
    unit = USD
    definition_period = YEAR
    reference = "https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-bundle.pdf#page=4"
    defined_for = StateCode.OH

    # def formula(tax_unit, period, parameters):
    #     return



from fiscalsim_us.model_api import *


class mt_standard_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Montana standard deduction"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        p = parameters(period).gov.states.mt.tax.income.deductions.standard
        agi = tax_unit("mt_agi", period)
        # standard deduction is a percentage of AGI that
        # is bounded by a min/max by filing status.
        min_amount = p.min[filing_status]
        max_amount = p.max[filing_status]
        uncapped_amount = p.rate * agi
        deduction_amount = min_(uncapped_amount, max_amount)

        return max_(deduction_amount, min_amount)



from fiscalsim_us.model_api import *


class mn_standard_deduction(Variable):
    """
    Line 4 of Minnesota 2022 Individual Income Tax return from M1.
    """

    value_type = float
    entity = TaxUnit
    label = "MN standard deduction"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MN

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.mn.tax.income.deductions.standard

        fed_agi = tax_unit("adjusted_gross_income", period)
        filing_status = tax_unit("filing_status", period)
        aged_blind_count = tax_unit("aged_blind_count", period)

        std_deduct = (
            p.base[filing_status] + p.extra[filing_status] * aged_blind_count
        )

        over_threshold = fed_agi >= p.threshold[filing_status]

        phaseout_subtract = min_(
            (fed_agi - p.threshold[filing_status]) * p.income_mult,
            std_deduct * p.mult,
        )
        return std_deduct - over_threshold * phaseout_subtract



from fiscalsim_us.model_api import *


class ebb(Variable):
    value_type = float
    entity = SPMUnit
    label = "Emergency Broadband Benefit amount"
    documentation = "Emergency Broadband Benefit amount"
    definition_period = YEAR
    unit = USD

    def formula(spm_unit, period, parameters):
        eligible = spm_unit("is_ebb_eligible", period)
        broadband_cost = spm_unit("broadband_cost_after_lifeline", period)
        tribal = spm_unit.household("is_on_tribal_land", period)
        amounts = parameters(period).gov.fcc.ebb.amount
        max_amount = (
            where(tribal, amounts.tribal, amounts.standard) * MONTHS_IN_YEAR
        )
        amount_if_eligible = min_(max_amount, broadband_cost)
        return where(eligible, amount_if_eligible, 0)



from fiscalsim_us.model_api import *


class dc_tanf_need_standard(Variable):
    value_type = float
    entity = SPMUnit
    label = "DC TANF need standard"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.DC

    def formula(spm_unit, period, parameters):
        unit_size = spm_unit("spm_unit_size", period)
        p = parameters(period).gov.states.dc.dhs.tanf.need_standard
        base = p.main[unit_size]
        # Add childcare addition, which depends on the child's age.
        person = spm_unit.members
        child = person("is_child", period)
        age = person("age", period)
        has_childcare_expenses = spm_unit("childcare_expenses", period) > 0
        # Look up supplement by age, and limit to children.
        person_childcare_supplement = p.additional_childcare.calc(age) * child
        # Aggregate person-level childcare supplement to SPM unit.
        spm_unit_childcare_supplement = has_childcare_expenses * spm_unit.sum(
            person_childcare_supplement
        )
        monthly = base + spm_unit_childcare_supplement
        return monthly * MONTHS_IN_YEAR



from fiscalsim_us.model_api import *


class sc_cdcc(Variable):
    """
    South Carolina Child and Dependent Care Credit, line 11 on form SC1040
    """

    value_type = float
    entity = TaxUnit
    label = "South Carolina CDCC"
    documentation = "South Carolina Child and Dependent Care Credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.sc.gov/forms-site/Forms/IITPacket_2023.pdf#page=32"
    )
    defined_for = StateCode.SC

    def formula(tax_unit, period, parameters):
        # Get South Carolina CDCC rate.
        p_sc = parameters(period).gov.states.sc.tax.income.credits.cdcc
        p_us = parameters(period).gov.irs.credits.cdcc

        # Married filing separate are ineligible.
        filing_status = tax_unit("filing_status", period)
        eligible = filing_status != filing_status.possible_values.SEPARATE

        # Number of qualifying people
        count_cdcc_eligible = min_(
            tax_unit("count_cdcc_eligible", period), p_us.eligibility.max
        )
        # Calculate the total state CDCC match for years starting before 2023
        pre_2023 = period.start.year < 2023

        # Year 2021 is different from federal cdcc
        max_decoupled_year_offset = p_sc.max_care_expense_year_offset
        period_max = period.offset(max_decoupled_year_offset)
        sc_max_care_expense = parameters(period_max).gov.irs.credits.cdcc.max

        # Get child care expenses.
        childcare_expenses = tax_unit("tax_unit_childcare_expenses", period)

        # Maximum value cannot exceed cap
        # Calculate total CDCC
        capped_expenses = min_(
            childcare_expenses, sc_max_care_expense * count_cdcc_eligible
        )
        sc_pre2023_tot_cdcc_match = eligible * capped_expenses * p_sc.rate

        # Calculate the total state CDCC match for years starting on or after
        # 2023
        max_expense = min_(
            p_sc.max_amount_pchild * count_cdcc_eligible, p_sc.total_max
        )
        sc_post2023_tot_cdcc_match = min_(
            max_expense, p_sc.rate * tax_unit("cdcc", period)
        )
        sc_tot_cdcc_match = (
            pre_2023 * sc_pre2023_tot_cdcc_match
            + (not pre_2023) * sc_post2023_tot_cdcc_match
        )
        return sc_tot_cdcc_match



from fiscalsim_us.model_api import *


class or_income_after_subtractions(Variable):
    value_type = float
    entity = TaxUnit
    label = "OR income after subtractions"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.OR
    adds = ["or_income_after_additions"]
    subtracts = ["or_income_subtractions"]



from fiscalsim_us.model_api import *
import numpy as np


class co_state_supplement_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Colorado State Supplement Eligible"
    definition_period = YEAR
    defined_for = StateCode.CO

    def formula(person, period, parameters):
        ssi_eligible = person("is_ssi_eligible_individual", period)
        is_disabled = person("is_ssi_disabled", period)
        is_blind = person("is_blind", period)
        age = person("age", period)
        p = parameters(period).gov.states.co.ssa.state_supplement

        # Convert to boolean arrays and handle potential None or NaN values
        ssi_eligible = np.array(ssi_eligible).astype(bool)
        is_disabled = np.array(is_disabled).astype(bool)
        is_blind = np.array(is_blind).astype(bool)

        disabled_or_blind = is_disabled | is_blind
        in_age_range = p.age_range.calc(age)

        # Convert in_age_range to boolean array
        in_age_range = np.array(in_age_range).astype(bool)

        # Use bitwise & for the final operation
        return disabled_or_blind & ssi_eligible & in_age_range



from fiscalsim_us.model_api import *


class ne_itemized_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "NE itemized deductions"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://revenue.nebraska.gov/files/doc/tax-forms/2021/f_1040n_booklet.pdf"
        "https://revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_5.pdf"
    )
    defined_for = StateCode.NE

    def formula(tax_unit, period, parameters):
        itm_deds_less_salt = tax_unit("itemized_deductions_less_salt", period)
        capped_property_taxes = tax_unit("capped_property_taxes", period)
        return itm_deds_less_salt + capped_property_taxes



from fiscalsim_us.model_api import *


class wi_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "Wisconsin exemption"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.wi.gov/TaxForms2021/2021-Form1f.pdf"
        "https://www.revenue.wi.gov/TaxForms2021/2021-Form1-Inst.pdf"
        "https://www.revenue.wi.gov/TaxForms2022/2022-Form1f.pdf"
        "https://www.revenue.wi.gov/TaxForms2022/2022-Form1-Inst.pdf"
        "https://docs.legis.wisconsin.gov/misc/lfb/informational_papers/january_2023/0002_individual_income_tax_informational_paper_2.pdf"
    )
    defined_for = StateCode.WI

    def formula(tax_unit, period, parameters):
        # compute base exemption amount
        p = parameters(period).gov.states.wi.tax.income
        base_amount = tax_unit("xtot", period) * p.exemption.base
        # compute extra exemption amount
        elderly_head = (
            tax_unit("age_head", period) >= p.exemption.old_age
        ).astype(int)
        elderly_spouse = (
            tax_unit("age_spouse", period) >= p.exemption.old_age
        ).astype(int)
        extra_amount = (elderly_head + elderly_spouse) * p.exemption.extra
        # return total exemption amount
        return base_amount + extra_amount



from fiscalsim_us.model_api import *


class refundable_ctc(Variable):
    value_type = float
    entity = TaxUnit
    label = "refundable CTC"
    unit = USD
    documentation = "The portion of the Child Tax Credit that is refundable."
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/24#d"

    def formula(tax_unit, period, parameters):
        # This line corresponds to "the credit which would be allowed under this section [the CTC section]"
        # without regard to this subsection [the refundability section] and the limitation under
        # section 26(a) [the section that limits the amount of the non-refundable CTC to tax liability].
        # This is the full CTC. This is then limited to the maximum refundable amount per child as per the
        # TCJA provision.

        ctc = parameters(period).gov.irs.credits.ctc

        maximum_amount = tax_unit("ctc_refundable_maximum", period)

        if ctc.refundable.fully_refundable:
            reduction = tax_unit("ctc_phase_out", period)
            return max_(0, maximum_amount - reduction)

        total_ctc = tax_unit("ctc", period)
        maximum_refundable_ctc = min_(maximum_amount, total_ctc)

        # The other part of the "lesser of" statement is: "the amount by which [the non-refundable CTC]
        # would increase if [tax liability] increased by tax_increase", where tax_increase is the greater of:
        # - the phase-in amount
        # - Social Security tax minus the EITC
        # First, we find tax_increase:

        earnings = tax_unit("tax_unit_earned_income", period)
        earnings_over_threshold = max_(
            0, earnings - ctc.refundable.phase_in.threshold
        )
        relevant_earnings = (
            earnings_over_threshold * ctc.refundable.phase_in.rate
        )

        # Compute "Social Security taxes" as defined in the US Code for the ACTC.
        # This includes OASDI and Medicare payroll taxes, as well as half
        # of self-employment taxes.
        SS_ADD_VARIABLES = [
            # Person:
            "employee_social_security_tax",
            "employee_medicare_tax",
            "unreported_payroll_tax",
            # Tax unit:
            "self_employment_tax_ald",
            "additional_medicare_tax",
        ]
        SS_SUBTRACT_VARIABLES = ["excess_payroll_tax_withheld"]
        social_security_tax = add(tax_unit, period, SS_ADD_VARIABLES) - add(
            tax_unit, period, SS_SUBTRACT_VARIABLES
        )
        eitc = tax_unit("eitc", period)
        social_security_excess = max_(0, social_security_tax - eitc)
        qualifying_children = tax_unit("ctc_qualifying_children", period)
        tax_increase = where(
            qualifying_children
            < ctc.refundable.phase_in.min_children_for_ss_taxes_minus_eitc,
            relevant_earnings,
            max_(relevant_earnings, social_security_excess),
        )
        limiting_tax = tax_unit("ctc_limiting_tax_liability", period)
        ctc_capped_by_tax = min_(total_ctc, limiting_tax)
        ctc_capped_by_increased_tax = min_(
            total_ctc, limiting_tax + tax_increase
        )
        amount_ctc_would_increase = (
            ctc_capped_by_increased_tax - ctc_capped_by_tax
        )

        return min_(maximum_refundable_ctc, amount_ctc_would_increase)


c11070 = variable_alias("c11070", refundable_ctc)



from fiscalsim_us.model_api import *


class ut_income_tax_before_credits(Variable):
    """
    Line 10 on Utah 2022 Individual Income Tax return form TC-40.
    """

    value_type = float
    entity = TaxUnit
    label = "Utah Income Tax Before Credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.UT

    def formula(tax_unit, period, parameters):
        rate = parameters(period).gov.states.ut.tax.income.rate
        income = tax_unit("ut_taxable_income", period)
        return max_(rate * income, 0)



from fiscalsim_us.model_api import *


class mn_intangible_drilling_costs(Variable):
    """
    Line 6 Minnesota 2022 form M1MT, Alternative Minimum Tax
    """

    value_type = float
    entity = TaxUnit
    label = "MN intangible drilling costs"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MN
